##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import bz2
import time
import zlib


from contextlib import closing

from zope import interface

from relstorage._util import get_memory_usage
from relstorage._util import byte_display
from relstorage._util import timer as _timer
from relstorage._util import log_timed as _log_timed
from relstorage._util import consume
from relstorage._compat import OID_TID_MAP_TYPE as OidTMap
from relstorage.interfaces import Int

from relstorage.cache.interfaces import IStateCache
from relstorage.cache.interfaces import IPersistentCache
from relstorage.cache.interfaces import IGenerationalLRUCache
from relstorage.cache.interfaces import IGeneration
from relstorage.cache.interfaces import ILRUEntry


from relstorage.cache.persistence import sqlite_connect
from relstorage.cache.persistence import sqlite_files
from relstorage.cache.persistence import FAILURE_TO_OPEN_DB_EXCEPTIONS
from relstorage.cache.local_database import Database

from relstorage.cache import cache

logger = __import__('logging').getLogger(__name__)

# pylint:disable=too-many-lines

class ICachedValue(ILRUEntry):
    """
    Data stored in the cache for a single OID.

    This may be a single ``(state, tid)`` pair, or it may be multiple
    such pairs, representing evolution of the object.

    Memory and time efficiency both matter. These objects do not know
    their own OID, just the state and tid.

    .. rubric:: Freezing objects

    For any given OID, one TID may be frozen. It may then be looked up
    without knowing its actual TID (using a TID of ``None``).
    This is useful for objects that do
    not change. Invalidations of frozen states happen automatically
    during the MVCC vacuuming process:

        - Freezing happens after a poll, when we have determined that
          an object has not changed within the range of transactions
          visible to all database viewers. The index entry is removed,
          and we begin looking for it at None.

          At this time, any older states cached for the object are removed.
          By definition, there can be no newer states, so only one state is
          accessible. (Of course, if we've just completed a transaction and
          not yet polled, then that's not strictly true; there could be
          cached data from the future not yet visible to any viewers.)

        - If an object previously frozen is changed, we see that in
          our index and won't ask for frozen states anymore.

          If we then load the new state from the DB, we cache it, leaving it
          with two cached states. Older viewers unaware of the change and accessing the database
          prior to it, can still use the frozen
          revision.

          Eventually that index entry reaches the end of its lifetime. If
          the object has not changed again, we will freeze it. This will
          discard the older frozen value and replace it with a new one. If
          it has changed again, we will invalidate the cache for anything
          older than that TID, which includes the first frozen state.

    The implementation and usage of frozen objects is contained entirely within
    this module. Clients are then responsible for making sure that the
    returned tid, if any, is within their viewing range.
    """

    # pylint:disable=no-self-argument,unexpected-special-method-signature,inherit-non-class
    # pylint:disable=arguments-differ

    weight = Int(
        description=u"""The cost (size) of this cached value.""")

    max_tid = Int(
        description=u"""The newest TID cached for the object."""
    )

    newest_value = interface.Attribute(u"The ``(state, tid)`` pair that is the newest.")

    def get_if_tid_matches(tid):
        """
        Return the ``(state, tid)`` for the given TID.

        A special value of None matches any previously frozen TID.

        If no TID matches, returns None.

        """

    def freeze_to_tid(tid):
        """
        Mark the given TID, if it exists, as frozen.

        Returns a new value to store in the cache. If it returns None,
        this entry is removed from the cache (because the TID didn't match,
        and must have been older.)
        """

    def with_later(value):
        """
        Add the ``(state, tid)`` (another ICachedValue) to the list of
        cached entries. Return the new value to place in the cache.
        """

    def discarding_tids_before(tid):
        """
        Remove the tid, and anything older, from the list of cached entries.

        Return the new value to place in the cache. Return None if all values
        were removed and the cached entry should be removed.
        """

interface.classImplements(cache.CachedValue, ICachedValue)
interface.classImplements(cache.PyCache, IGenerationalLRUCache)
interface.classImplements(cache.PyGeneration, IGeneration)


@interface.implementer(IStateCache,
                       IPersistentCache)
class LocalClient(object):
    # pylint:disable=too-many-public-methods,too-many-instance-attributes

    # Percentage of our byte limit that should be dedicated
    # to the main "protected" generation
    _gen_protected_pct = 0.8
    # Percentage of our byte limit that should be dedicated
    # to the initial "eden" generation
    _gen_eden_pct = 0.1
    # Percentage of our byte limit that should be dedicated
    # to the "probationary"generation
    _gen_probation_pct = 0.1
    # By default these numbers add up to 1.0, but it would be possible to
    # overcommit by making them sum to more than 1.0. (For very small
    # limits, the rounding will also make them overcommit).

    # Use the same markers as zc.zlibstorage (well, one marker)
    # to automatically avoid double-compression
    _compression_markers = {
        'zlib': (b'.z', zlib.compress),
        'bz2': (b'.b', bz2.compress),
        'none': (None, None)
    }
    _decompression_functions = {
        b'.z': zlib.decompress,
        b'.b': bz2.decompress
    }

    # What multiplier of the number of items in the cache do we apply
    # to determine when to age the frequencies?
    _age_factor = 10

    # When did we last age?
    _aged_at = 0
    _next_age_at = 1000

    _cache = None

    # Things copied from self._cache
    _peek = None

    def __init__(self, options,
                 prefix=None):
        self.options = options
        self.prefix = prefix or ''
        # XXX: The calc for limit is substantially smaller
        # The real MB value is 1024 * 1024 = 1048576
        self.limit = int(1000000 * options.cache_local_mb)
        self._value_limit = options.cache_local_object_max

        # The underlying data storage. It maps ``{oid: value}``,
        # where ``value`` is an :class:`ICachedValue`.
        #
        # Keying off of OID directly lets the LRU be based on the
        # popularity of an object, not its particular transaction. It also lets
        # us use a LLBTree to store the data, which can be more memory efficient.
        self._cache = None

        self.flush_all()
        self.__initial_weight = self._cache.weight

        compression_module = options.cache_local_compression
        try:
            compression_markers = self._compression_markers[compression_module]
        except KeyError:
            raise ValueError("Unknown compression module")
        else:
            self.__compression_marker = compression_markers[0]
            self.__compress = compression_markers[1]
            if self.__compress is None:
                self._compress = None

    @property
    def size(self):
        return self._cache.weight

    def __len__(self):
        """
        How many distinct OIDs are stored.
        """
        return len(self._cache)

    def __iter__(self):
        return iter(self._cache)

    def keys(self):
        return self._cache.keys()

    def _decompress(self, data):
        pfx = data[:2]
        if pfx not in self._decompression_functions:
            return data
        return self._decompression_functions[pfx](data[2:])

    def _compress(self, data): # pylint:disable=method-hidden
        # We override this if we're disabling compression
        # altogether.
        # Use the same basic rule as zc.zlibstorage, but bump the object size up from 20;
        # many smaller object (under 100 bytes) like you get with small btrees,
        # tend not to compress well, so don't bother.
        if data and (len(data) > 100) and data[:2] not in self._decompression_functions:
            compressed = self.__compression_marker + self.__compress(data)
            if len(compressed) < len(data):
                return compressed
        return data

    @_log_timed
    def save(self, object_index=None, checkpoints=None, **sqlite_args):
        options = self.options
        if options.cache_local_dir and self.size > self.__initial_weight:
            try:
                conn = sqlite_connect(options, self.prefix,
                                      **sqlite_args)
            except FAILURE_TO_OPEN_DB_EXCEPTIONS:
                logger.exception("Failed to open sqlite to write")
                return 0

            with closing(conn):
                self.write_to_sqlite(conn, checkpoints, object_index)
            # Testing: Return a signal when we tried to write
            # something.
            return 1

    def restore(self):
        """
        Load the data from the persistent database.

        Returns the checkpoint data last saved, which may be None if
        there was no data.
        """
        options = self.options
        if options.cache_local_dir:
            try:
                conn = sqlite_connect(options, self.prefix)
            except FAILURE_TO_OPEN_DB_EXCEPTIONS:
                logger.exception("Failed to read data from sqlite")
                return
            with closing(conn):
                return self.read_from_sqlite(conn)

    @_log_timed
    def remove_invalid_persistent_oids(self, bad_oids):
        """
        Remove data from the persistent cache for the given oids.
        """
        options = self.options
        if not options.cache_local_dir:
            return

        count_removed = 0
        conn = '(no oids to remove)'
        if bad_oids:
            self.invalidate_all(bad_oids)
            conn = sqlite_connect(options, self.prefix)
            with closing(conn):
                db = Database.from_connection(conn)
                count_removed = db.remove_invalid_persistent_oids(bad_oids)

        logger.debug("Removed %d invalid OIDs from %s", count_removed, conn)

    def zap_all(self):
        _, destroy = sqlite_files(self.options, self.prefix)
        destroy()
        # zapping happens frequently during test runs,
        # and during zodbconvert when the process will exist
        # only for a short time.
        self.flush_all()

    def flush_all(self):
        if self._cache or self._cache is None:
            # Only actually abandon the cache object
            # if it has data. Otherwise let it keep the
            # preallocated CFFI objects it may have
            # (those are expensive to create and tests call
            # this a LOT)
            byte_limit = self.limit
            self._cache = cache.PyCache(
                byte_limit * self._gen_eden_pct,
                byte_limit * self._gen_protected_pct,
                byte_limit * self._gen_probation_pct
            )
        self._peek = self._cache.peek
        self.reset_stats()

    def reset_stats(self):
        self._cache.reset_stats()
        self._aged_at = 0
        self._next_age_at = 1000

    def stats(self):
        total = self._cache.hits + self._cache.misses
        return {
            'hits': self._cache.hits,
            'misses': self._cache.misses,
            'sets': self._cache.sets,
            'ratio': self._cache.hits / total if total else 0,
            'len': len(self),
            'bytes': self.size,
        }

    def __contains__(self, oid_tid):
        oid, tid = oid_tid
        return self._cache.get_item_with_tid(oid, tid) is not None

    def get(self, oid_tid, peek=False):
        oid, tid = oid_tid
        assert tid is None or tid >= 0
        decompress = self._decompress
        value = None

        if peek:
            value = self._cache.peek_item_with_tid(oid, tid)
        else:
            value = self._cache.get_item_with_tid(oid, tid)

        # Finally, decompress if needed.
        # Recall that for deleted objects, `state` can be None.
        # TODO: This is making a copy of the string from C++ even if we
        # don't need to decompress.
        if value is not None:
            state, tid = value
            return ((decompress(state) if state else state), tid)

    __getitem__ = get

    def _age(self):
        # Age only when we're full and would thus need to evict; this
        # makes initial population faster. It's cheaper to calculate this
        # AFTER the operations, though, because we read it from C.
        #if self.size < self.limit:
        #    return

        # Age the whole thing periodically based on the number of
        # operations we've done that would have altered popularity.
        # Dynamically calculate how often we need to age. By default, this is
        # based on what Caffeine's PerfectFrequency does: 10 * max
        # cache entries.
        #
        # We don't take a lock to do this; it's fine if two threads
        # attempt it at the same time.
        age_period = self._age_factor * len(self._cache)
        operations = self._cache.hits + self._cache.sets
        if operations - self._aged_at < age_period:
            self._next_age_at = age_period
            return
        if self.size < self.limit:
            return

        self._aged_at = operations
        now = time.time()
        logger.debug("Beginning frequency aging for %d cache entries",
                     len(self._cache))
        self._cache.age_frequencies()
        done = time.time()
        logger.debug("Aged %d cache entries in %s", len(self._cache), done - now)

        self._next_age_at = int(self._aged_at * 1.5) # in case the dict shrinks

        return self._aged_at

    def __setitem__(self, oid_tid, state_bytes_tid):
        if not self.limit:
            # don't bother
            return

        # A state of 'None' happens for undone transactions.
        oid, key_tid = oid_tid
        state_bytes, actual_tid = state_bytes_tid

        # Really key_tid should be > 0; we allow >= for tests.
        assert key_tid == actual_tid and key_tid >= 0
        self.set_all_for_tid(key_tid, [(state_bytes, oid, None)])

    def set_all_for_tid(self, tid_int, state_oid_iter):
        if self.limit:
            self._cache.set_all_for_tid(tid_int, state_oid_iter, self._compress, self._value_limit)
            # Inline some of the logic about whether to age or not; avoiding the
            # call helps speed
            if self._cache.hits + self._cache.sets > self._next_age_at:
                self._age()

    def __delitem__(self, oid_tid):
        self.delitems({oid_tid[0]: oid_tid[1]})

    def delitems(self, oids_tids):
        """
        For each OID/TID pair in the items, remove all cached values
        for OID that are older than TID.
        """
        self._cache.delitems(oids_tids)

    def invalidate_all(self, oids):
        self._cache.del_oids(oids)

    def freeze(self, oids_tids):
        self._cache.freeze(oids_tids)

    def close(self):
        pass

    release = close

    def new_instance(self):
        return self

    def updating_delta_map(self, deltas):
        return deltas

    def _bulk_update(self, keys_and_values,
                     source='<unknown>',
                     log_count=None,
                     mem_usage_before=None):
        """
        Insert all the
        ``(oid, (state, actual_tid, frozen, frequency)))`` pairs
        found in *keys_and_values*, rejecting entries once we get too full.

        *keys_and_values* should be ordered from least to most recently used.

        This can only be done in an empty cache.
        """
        now = time.time()
        mem_usage_before = mem_usage_before if mem_usage_before is not None else get_memory_usage()
        mem_usage_before_this = get_memory_usage()
        log_count = log_count or len(keys_and_values)

        data = self._cache
        if data:
            raise ValueError("Only bulk load into an empty cache.")

        stored = data.add_MRUs(
            keys_and_values,
            return_count_only=True)

        then = time.time()
        del keys_and_values # For memory reporting.
        mem_usage_after = get_memory_usage()
        logger.info(
            "Examined %d and stored %d items from %s in %s using %s. "
            "(%s local) (%s)",
            log_count, stored, getattr(source, 'name', source),
            then - now,
            byte_display(mem_usage_after - mem_usage_before),
            byte_display(mem_usage_after - mem_usage_before_this),
            self)
        return log_count, stored


    @_log_timed
    def read_from_sqlite(self, connection):
        import gc
        gc.collect() # Free memory, we're about to make big allocations.
        mem_before = get_memory_usage()

        db = Database.from_connection(connection)
        checkpoints = db.checkpoints

        @_log_timed
        def fetch_and_filter_rows():
            # Do this here so that we don't have the result
            # in a local variable and it can be collected
            # before we measure the memory delta.
            #
            # In large benchmarks, this function accounts for 57%
            # of the total time to load data. 26% of the total is
            # fetching rows from sqlite, and 18% of the total is allocating
            # storage for the blob state.
            #
            # We make one call into sqlite and let it handle the iterating.
            # Items are (oid, key_tid, state, actual_tid).
            # key_tid may equal the actual tid, or be -1 when the row was previously
            # frozen;
            # That doesn't matter to us, we always freeze all rows.
            size = 0
            limit = self.limit
            items = []
            rows = db.fetch_rows_by_priority()
            for oid, frozen, state, actual_tid, frequency in rows:
                size += len(state)
                if size > limit:
                    break
                items.append((oid, (state, actual_tid, frozen, frequency)))
            consume(rows)
            # Rows came to us MRU to LRU, but we need to feed them the other way.
            items.reverse()
            return items

        # In the large benchmark, this is 25% of the total time.
        # 18% of the total time is preallocating the entry nodes.
        self._bulk_update(fetch_and_filter_rows(),
                          source=connection,
                          mem_usage_before=mem_before)
        return checkpoints

    def _items_to_write(self, stored_oid_tid):
        # pylint:disable=too-many-locals
        all_entries_len = len(self._cache)

        # Only write the newest entry for each OID.


        # Newly added items have a frequency of 1. They *may* be
        # useless stuff we picked up from an old cache file and
        # haven't touched again, or they may be something that
        # really is new and we have no track history for (that
        # would be in eden generation).
        #
        # Ideally we want to do something here that's similar to what
        # the SegmentedLRU does: if we would reject an item from eden,
        # then only keep it if it's better than the least popular
        # thing in probation.
        #
        # It's not clear that we have a good algorithm for this, and
        # what we tried takes a lot of time, so we don't bother.
        # TODO: Finish tuning these.

        # But we have a problem: If there's an object in the existing
        # cache that we have updated but now exclude, we would fail to
        # write its new data. So we can't exclude anything that's
        # changed without also nuking it from the DB.

        # Also, if we had a value loaded from the DB, and then we
        # modified it, but later ejected all entries for that object
        # from the cache, we wouldn't see it here, and we could be
        # left with stale data in the database. We must track the
        # (oid, tid) we load from the database and record that we
        # should DELETE those rows if that happens.
        # This is done with the __min_allowed_writeback dictionary,
        # which we mutate here: When we find a row we can write that meets
        # the requirement, we take it out of the dictionary.
        # Later, if there's anything left in the dictionary, we *know* there
        # may be stale entries in the cache, and we remove them.

        # The *object_index* is our best polling data; anything it has it gospel,
        # so if it has an entry for an object, it superceeds our own.
        get_current_oid_tid = stored_oid_tid.get
        matching_tid_count = 0
        # When we accumulate all the rows here before returning them,
        # this function shows as about 3% of the total time to save
        # in a very large database.
        with _timer() as t:
            for oid, lru_entry in self._cache.iteritems():
                newest_value = lru_entry.newest_value
                # We must have something at least this fresh
                # to consider writing it out
                if newest_value is None:
                    raise AssertionError("Value should not be none", oid, lru_entry)
                actual_tid = newest_value.tid

                # If we have something >= min_allowed, matching
                # what's in the database, or even older (somehow),
                # it's not worth writing to the database (states should be identical).
                current_tid = get_current_oid_tid(oid, -1)
                if current_tid >= actual_tid:
                    matching_tid_count += 1
                    continue

                yield (oid, actual_tid, newest_value.frozen,
                       bytes(newest_value.state),
                       lru_entry.frequency)

                # We're able to satisfy this, so we don't need to consider
                # it in our min_allowed set anymore.
                # XXX: This isn't really the right place to do this.
                # Move this and write a test.


        removed_entry_count = matching_tid_count
        logger.info(
            "Examined %d entries and rejected %d "
            "(already in db: %d) in %s",
            all_entries_len, removed_entry_count,
            matching_tid_count,
            t.duration)

    @_log_timed
    def write_to_sqlite(self, connection, checkpoints, object_index=None):
        # pylint:disable=too-many-locals
        mem_before = get_memory_usage()
        object_index = object_index or OidTMap()

        cur = connection.cursor()

        db = Database.from_connection(connection)
        begin = time.time()

        # In a large benchmark, store_temp() accounts for 32%
        # of the total time, while move_from_temp accounts for 49%.

        # The batch size depends on how many params a stored proc can
        # have; if we go too big we get OperationalError: too many SQL
        # variables. The default is 999.
        # Note that the multiple-value syntax was added in
        # 3.7.11, 2012-03-20.
        with _timer() as batch_timer:
            cur.execute('BEGIN')
            stored_oid_tid = db.oid_to_tid
            fetch_current = time.time()
            count_written, _ = db.store_temp(self._items_to_write(stored_oid_tid))
            cur.execute("COMMIT")


        # Take out (reserved write) locks now; if we don't, we can get
        # 'OperationalError: database is locked' (the SQLITE_BUSY
        # error code; the SQLITE_LOCKED error code is about table
        # locks and has the string 'database table is locked') But
        # beginning immediate lets us stand in line.
        #
        # Note that the SQLITE_BUSY error on COMMIT happens "if an
        # another thread or process has a shared lock on the database
        # that prevented the database from being updated.", But "When
        # COMMIT fails in this way, the transaction remains active and
        # the COMMIT can be retried later after the reader has had a
        # chance to clear." So we could retry the commit a couple
        # times and give up if can't get there. It looks like one
        # production database takes about 2.5s to execute this entire function;
        # it seems like most instances that are shutting down get to this place
        # at roughly the same time and stack up waiting:
        #
        # Instance | Save Time | write_to_sqlite time
        #     1    |   2.36s   |   2.35s
        #     2    |   5.31s   |   5.30s
        #     3    |   6.51s   |   6.30s
        #     4    |   7.91s   |   7.90s
        #
        # That either suggests that all waiting happens just to acquire this lock and
        # commit this transaction (and that subsequent operations are inconsequential
        # in terms of time) OR that the exclusive lock isn't truly getting dropped,
        # OR that some function in subsequent processes is taking longer and longer.
        # And indeed, our logging showed that a gc.collect() at the end of this function was
        # taking more and more time:
        #
        # Instance | GC Time  | Reported memory usage after GC
        #    1     |   2.00s  |     34.4MB
        #    2     |   3.50s  |     83.2MB
        #    3     |   4.94s  |    125.9MB
        #    4     |   6.44   |    202.1MB
        #
        # The exclusive lock time did vary, but not as much; trim time
        # didn't really vary:
        #
        # Instance |  Exclusive Write Time | Batch Insert Time | Fetch Current
        #    1     |   0.09s               |   0.12s           |   0.008s
        #    2     |   1.19s               |   0.44s           |   0.008s
        #    3     |   0.91s               |   0.43s           |   0.026s
        #    4     |   0.92s               |   0.34s           |   0.026s
        with _timer() as exclusive_timer:
            cur.execute('BEGIN IMMEDIATE')
            # During the time it took for us to commit, and the time that we
            # got the lock, it's possible that someone else committed and
            # changed the data in object_state
            # Things might have changed in the database since our
            # snapshot where we put temp objects in, so we can't rely on
            # just assuming that's the truth anymore.
            rows_inserted = db.move_from_temp()
            if checkpoints:
                db.update_checkpoints(*checkpoints)

            cur.execute('COMMIT')
        # TODO: Maybe use BTrees.family.intersection to get the common keys?
        # Or maybe use a SQLite Python function?
        # 'DELETE from object_state where is_stale(zoid, tid)'
        # What's fastest? The custom function version would require scanning the entire
        # table; unfortunately this sqlite interface doesn't allow defining virtual tables.
        with _timer() as trim_timer:
            # Delete anything we still know is stale, because we
            # saw a new TID for it, but we had nothing to replace it with.
            min_allowed_writeback = OidTMap()
            db.trim_to_size(self.limit, min_allowed_writeback)
            del min_allowed_writeback

        # Cleanups
        cur.close()
        del cur
        db.close() # closes the connection.
        del db
        del stored_oid_tid

        # We're probably shutting down, don't perform a GC; we see
        # that can get quite lengthy.

        end = time.time()
        stats = self.stats()
        mem_after = get_memory_usage()
        logger.info(
            "Wrote %d items to %s in %s "
            "(%s to fetch current, %s to insert batch (%d rows), %s to write, %s to trim). "
            "Used %s memory. (Storage: %s) "
            "Total hits %s; misses %s; ratio %s (stores: %s)",
            count_written, connection, end - begin,
            fetch_current - begin, batch_timer.duration, rows_inserted, exclusive_timer.duration,
            trim_timer.duration,
            byte_display(mem_after - mem_before), self._cache,
            stats['hits'], stats['misses'], stats['ratio'], stats['sets'])

        return count_written



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from hamcrest import assert_that
from nti.testing.matchers import validly_provides

from relstorage.tests import TestCase
from relstorage.tests import MockAdapter
from relstorage.options import Options
from relstorage.cache import interfaces
from relstorage.cache import mvcc

from . import LocalClient

class MockObjectIndex(object):

    maps = ()
    detached = False
    highest_visible_tid = None
    complete_since_tid = None

    def __setitem__(self, k, v):
        raise AttributeError

class MockViewer(object):

    object_index = None # type: MockObjectIndex
    highest_visible_tid = None
    detached = False

    def __init__(self, options=None):
        options = options or Options()
        self.adapter = MockAdapter()
        self.local_client = LocalClient(options)

    def __repr__(self):
        return "<MockViewer at 0x%x index=%r>" % (
            id(self),
            self.object_index,
        )

class TestMVCCDatabaseCorrdinator(TestCase):

    expected_poll_conn = None
    expected_poll_cursor = None
    expected_poll_last_tid = None
    expected_poll_result = None

    def setUp(self):
        self.coord = self._makeOne()
        self.viewer = MockViewer()
        self.coord.register(self.viewer)
        self.polled_tid = 0
        self.polled_changes = None
        self.viewer.adapter.poller.poll_invalidations = self.poll_invalidations
        self.viewer.adapter.poller.get_current_tid = self.get_current_tid

    def get_current_tid(self, _cursor):
        return self.polled_tid

    def poll_invalidations(self, conn, cursor, last_tid):
        self.assertEqual(last_tid, self.expected_poll_last_tid)
        self.assertEqual(conn, self.expected_poll_conn)
        self.assertEqual(cursor, self.expected_poll_cursor)
        return self.polled_changes, self.polled_tid

    def add_viewer(self):
        viewer = MockViewer()
        viewer.adapter = self.viewer.adapter
        self.coord.register(viewer)
        return viewer

    def tearDown(self):
        del self.viewer.adapter.poller.poll_invalidations
        self.coord.unregister(self.viewer)
        self.viewer = None
        self.coord = None

    def _makeOne(self):
        return mvcc.MVCCDatabaseCoordinator()

    def test_implements(self):
        assert_that(self._makeOne(),
                    validly_provides(interfaces.IStorageCacheMVCCDatabaseCoordinator))

    def assertNoPollingState(self):
        self.none(self.coord.object_index)
        self.none(self.viewer.object_index)
        self.none(self.viewer.highest_visible_tid)
        self.none(self.coord.maximum_highest_visible_tid)

    def do_poll(self, viewer=None):
        viewer = viewer or self.viewer
        result = self.coord.poll(viewer, None, None)
        if result:
            result = list(result)

        self.assertEqual(result, self.expected_poll_result)
        if self.polled_tid:
            # Only change if we expect a different poll to come in;
            # polling with 0 shouldn't change.
            self.expected_poll_last_tid = self.polled_tid

    def test_poll_no_index_when_0(self):
        # Starting from blank state, we only begin making an index
        # when there's data in the DB.
        cache = self.viewer
        cache.object_index = MockObjectIndex()
        coord = self.coord
        self.assertIsNone(coord.object_index)
        assert (self.polled_tid, self.polled_changes) == (0, None)
        # As long as it keeps returning 0, we don't try to begin caching
        for _ in range(2):
            self.do_poll()
            self.assertIsNone(coord.object_index)

        self.assertNoPollingState()

    def test_poll_no_index_begins(self, polled_tid=1):
        cache = self.viewer
        coord = self.coord

        self.polled_tid = polled_tid
        self.do_poll()
        self.expected_poll_last_tid = self.polled_tid

        self.assertIsNotNone(coord.object_index)
        self.assertIs(cache.object_index, coord.object_index)
        self.assertEqual(polled_tid, cache.object_index.highest_visible_tid)
        self.assertEqual(polled_tid, coord.maximum_highest_visible_tid)
        self.assertEqual(polled_tid, coord.minimum_highest_visible_tid)

    def test_tid_goes_0_after_begin(self):
        self.test_poll_no_index_begins()
        # zap the database
        self.polled_tid = 0

        self.do_poll()

        # We lost all our state
        self.assertNoPollingState()

    def test_tid_goes_back_after_begin(self):
        self.test_poll_no_index_begins(polled_tid=2)
        second_viewer = self.add_viewer()
        second_viewer.object_index = MockObjectIndex()


        # switch to a replica
        self.polled_tid = self.polled_tid - 1
        assert self.polled_tid == 1
        self.do_poll()
        self.assertNoPollingState()

        # Other registered viewers were marked invalid.
        self.assertTrue(second_viewer.detached)

        # Poll the first guy again, begin getting state.
        self.expected_poll_last_tid = None
        self.do_poll()
        self.assertIs(self.viewer.object_index, self.coord.object_index)
        self.assertIsNotNone(self.viewer.object_index)

        # And for the still-invalid viewer, polling continues from
        # where we just polled.
        self.expected_poll_last_tid = 1
        self.polled_changes = []
        self.expected_poll_result = None
        second_viewer.adapter = self.viewer.adapter
        self.do_poll(viewer=second_viewer)
        self.assertIs(second_viewer.object_index, self.viewer.object_index)

    def test_poll_many_times_vacuums_one_viewer(self):
        # Only one viewer, so state management is simple.

        # Start us off.
        self.test_poll_no_index_begins(2)

        # Put something in here that can get frozen.
        # Before we began polling.
        self.assertEqual(self.viewer.object_index.highest_visible_tid, 2)
        self.viewer.local_client[(0, 1)] = (b'cached data', 1)
        self.viewer.object_index[0] = 1

        # Move forward
        for poll_num in range(3, 20):
            __traceback_info__ = poll_num
            prev_poll = self.expected_poll_last_tid
            oid = poll_num
            tid = poll_num
            self.viewer.local_client[(oid, 1)] = (b'cached data', 1)
            self.assertIn((oid, 1), self.viewer.local_client)
            self.viewer.object_index[oid] = 1
            self.polled_tid = poll_num
            self.polled_changes = self.expected_poll_result = [(oid, tid)]
            self.do_poll()
            self.assertLength(self.viewer.object_index.maps, 1)
            self.assertEqual(self.coord.minimum_highest_visible_tid, poll_num)
            self.assertEqual(self.coord.maximum_highest_visible_tid, poll_num)
            # Our completion is our last poll.
            self.assertEqual(self.viewer.object_index.complete_since_tid, prev_poll)

        # All of the OIDs I put in the local client got invalidated
        # as we went, except for the first one, which got frozen
        # while remaining accessible at the old key without actually increasing
        # the number of entries in the cache or its byte size.
        self.assertEqual(len(self.viewer.local_client), 1)
        self.assertIn((0, None), self.viewer.local_client)
        self.assertIn((0, 1), self.viewer.local_client)


    def test_poll_many_times_vacuums_two_viewer(self):
        # A viewer that keeps moving forward, and a viewer that
        # is stuck in the past.
        second_viewer = self.add_viewer()

        self.test_poll_no_index_begins()
        # Second guy will actually ask for changes;
        # give him the same TID.
        self.polled_changes = ()
        self.expected_poll_result = None
        self.do_poll(viewer=second_viewer)

        # They have the same index to start with.
        self.assertIs(self.viewer.object_index, second_viewer.object_index)

        # Now move forward; the old viewer that's not updating his
        # state is keeping us from
        for poll_num in range(2, 20):
            __traceback_info__ = poll_num
            oid = poll_num
            tid = poll_num
            self.polled_tid = poll_num
            self.polled_changes = self.expected_poll_result = [(oid, tid)]
            self.do_poll()
            self.assertLength(self.viewer.object_index.maps, poll_num)
            # The max keeps going up.
            self.assertEqual(self.coord.maximum_highest_visible_tid, poll_num)
            # The min stays pinned
            self.assertEqual(self.coord.minimum_highest_visible_tid,
                             second_viewer.object_index.highest_visible_tid)
            self.assertEqual(self.viewer.object_index.complete_since_tid, 1)

        # At the end, the index is not the same.
        self.assertIsNot(self.viewer.object_index, second_viewer.object_index)
        # But the original map still is.
        self.assertIs(self.viewer.object_index.maps[-1],
                      second_viewer.object_index.maps[0])

    def test_poll_many_times_vacuums_several_viewers(self):
        # A viewer that keeps moving forward, and a viewer that
        # is stuck in the past.
        second_viewer = self.add_viewer()

        self.test_poll_no_index_begins()
        # Second guy will actually ask for changes;
        # give him the same TID. Because its his first time,
        # though, we return no changes to him.
        self.polled_changes = ()
        self.expected_poll_result = None
        self.do_poll(viewer=second_viewer)

        # Third guy exists, hasn't actually
        # viewed anything yet.
        third_viewer = self.add_viewer()

        # They have the same index to start with.
        self.assertIs(self.viewer.object_index, second_viewer.object_index)
        self.none(third_viewer.object_index)

        # Now move forward; the old viewer that's not updating his
        # state is keeping us from
        def poll_for_poll_num(poll_num):
            __traceback_info__ = poll_num
            oid = poll_num
            tid = poll_num
            self.polled_tid = poll_num
            self.polled_changes = self.expected_poll_result = [(oid, tid)]
            self.do_poll()
            self.assertLength(self.viewer.object_index.maps, poll_num)
            # The max keeps going up.
            self.assertEqual(self.coord.maximum_highest_visible_tid, poll_num)
            # The min stays pinned
            self.assertEqual(self.coord.minimum_highest_visible_tid,
                             second_viewer.object_index.highest_visible_tid)
            self.assertEqual(self.viewer.object_index.complete_since_tid, 1)

        for poll_num in range(2, 20):

            poll_for_poll_num(poll_num)

        # Now plop the third viewer down right in the middle of this sequence
        # and then go again.
        self.polled_tid = poll_num
        self.polled_changes = ()
        self.expected_poll_result = None
        self.do_poll(viewer=third_viewer)
        self.assertIs(self.viewer.object_index, third_viewer.object_index)
        for poll_num in range(20, 30):
            poll_for_poll_num(poll_num)

        # All three share the original map
        self.assertIs(self.viewer.object_index.maps[-1],
                      second_viewer.object_index.maps[0])
        self.assertIs(third_viewer.object_index.maps[-1],
                      second_viewer.object_index.maps[0])
        # The second and third share the back 20 maps
        new_maps = self.viewer.object_index.maps[10:]
        back_maps = third_viewer.object_index.maps[:]
        self.assertEqual(new_maps, back_maps)

        # Change settings: limit the depth
        self.coord.max_allowed_index_depth = 2
        # Poll; this throws off the oldest reader.
        poll_num += 1
        self.polled_tid = oid = tid = poll_num
        self.polled_changes = self.expected_poll_result = [(oid, tid)]
        self.do_poll()
        # The oldest reader is now invalid
        self.assertTrue(second_viewer.detached)
        # And the maps have been combined, back to the third_viewer
        # 12 = third_viewer + 10 intermediates + most recent poll
        self.assertLength(self.viewer.object_index.maps, 12)

        # polling again will drop the third viewer and shrink everything up.
        poll_num += 1
        self.polled_tid = oid = tid = poll_num
        self.polled_changes = self.expected_poll_result = [(oid, tid)]
        self.do_poll()

        self.assertTrue(third_viewer.detached)
        self.assertLength(self.viewer.object_index.maps, 1)

    def test_poll_produces_gap_vacuum(self):
        # TIDs do not always strictly go up; if there's a delay polling,
        # one viewer may not poll as far forward as everyone else and so
        # will produce a divergent index. vacuum need sto handle that.

        self.test_poll_no_index_begins()

        second_viewer = self.add_viewer()
        # Second guy will actually ask for changes;
        # give him the same TID. Because its his first time,
        # though, we return no changes to him.
        self.polled_changes = ()
        self.expected_poll_result = None
        self.do_poll(viewer=second_viewer)

        # Both now have the same min
        self.assertEqual(self.viewer.highest_visible_tid, second_viewer.highest_visible_tid)
        self.assertEqual(self.viewer.highest_visible_tid, self.coord.minimum_highest_visible_tid)
        orig_index = self.coord.object_index

        # Pull the first one out ahead
        changes_5 = (1, 5)
        changes_10 = (2, 10)
        self.polled_tid = 10
        self.polled_changes = self.expected_poll_result = [changes_5, changes_10]
        self.do_poll()
        index_10 = self.coord.object_index
        self.assertIsNot(orig_index, index_10)
        self.assertEqual(1, self.coord.minimum_highest_visible_tid)

        # Pull the other one to the middle, gapping the index.
        # In order to do this, we have to pretend that we started with
        # the old index and got interrupted before we finished.
        self.coord.object_index = orig_index
        self.expected_poll_last_tid = orig_index.minimum_highest_visible_tid
        self.polled_tid = 5
        self.polled_changes = self.expected_poll_result = [changes_5]
        self.do_poll(viewer=second_viewer)

        # Now the first one goes ahead and polls farther to the future.
        # Vacuum happens, but doesn't remove the original transaction index
        # we still need.
        self.coord.object_index = index_10
        self.assertEqual(self.coord.minimum_highest_visible_tid, 5)
        self.assertEqual(self.coord.object_index.depth, 2)
        self.expected_poll_last_tid = 10
        self.polled_tid = 15
        self.polled_changes = self.expected_poll_result = [(1, 15)]
        self.do_poll()
        # We vacuumed as far forward as we could.
        self.assertEqual(self.coord.minimum_highest_visible_tid, 5)
        self.assertEqual(self.coord.object_index.depth, 3)

    def test_restore_timeout(self):
        from relstorage.tests import mock
        from relstorage.tests import MockOptions
        from relstorage.adapters.mover import AbstractObjectMover
        from relstorage.adapters.batch import RowBatcher
        from relstorage.adapters.interfaces import AggregateOperationTimeoutError

        mock_perf_counter = mock.Mock()
        mock_perf_counter.side_effect = (
            12345,
            12346,
            12347
        )

        # We'll poll for 10 oids, all of which will be present
        oids = (1, 2, 3, 4, 5, 6, 7, 8, 9, 10)
        def configure_cursor(cur):
            cur.many_results = [
                [(oid, 1)]
                for oid in oids
            ]
            return cur
        # But we'll timeout after two batches,
        # and the batches will be of size 1 (for simplicity)
        timeout = 2
        batch_size = 1

        class MockRowBatcher(RowBatcher):
            ex = None
            def select_from(self, *args, **kwargs):
                # pylint:disable=signature-differs
                try:
                    for x in RowBatcher.select_from(self, *args, **kwargs):
                        yield x
                except AggregateOperationTimeoutError as ex:
                    MockRowBatcher.ex = ex
                    raise

        def batcher_factory(*args):
            batch = MockRowBatcher(*args)
            batch.bind_limit = batch_size
            batch.perf_counter = mock_perf_counter
            return batch


        adapter = MockAdapter()
        adapter.mover = AbstractObjectMover(None, MockOptions(), batcher_factory=batcher_factory)

        adapter.connmanager.configure_cursor = configure_cursor

        class MockCache(object):
            def contains_oid_with_tid(self, oid, tid):
                return oid is not None and tid is not None

        class MockLocalClient(object):
            _cache = MockCache()
            invalid_oids = None
            def restore(self):
                return (1, 1)

            def keys(self):
                return oids

            def remove_invalid_persistent_oids(self, oids):
                self.invalid_oids = list(oids)

        local_client = MockLocalClient()

        coord = self._makeOne()
        coord.restore(adapter, local_client, timeout=timeout)

        self.assertIsNotNone(MockRowBatcher.ex)
        self.assertEqual(
            dict(MockRowBatcher.ex.partial_result),
            {1: 1, 2: 1}
        )
        # Everything we didn't get returned from the database was
        # considered invalid.
        self.assertEqual(local_client.invalid_oids, [
            3, 4, 5, 6, 7, 8, 9, 10
        ])

class TestTransactionRangeObjectIndex(TestCase):

    def _makeOne(self,
                 highest_visible_tid,
                 complete_since_tid,
                 data):
        return mvcc._TransactionRangeObjectIndex(
            highest_visible_tid,
            complete_since_tid,
            data)

    def test_bad_tid_in_ctor(self):
        with self.assertRaises(AssertionError):
            self._makeOne(highest_visible_tid=1, complete_since_tid=2, data=())

    def test_bad_data(self):
        # Too high
        with self.assertRaises(AssertionError):
            self._makeOne(highest_visible_tid=2,
                          complete_since_tid=0,
                          data=[(1, 3)])

        # Too low
        with self.assertRaises(AssertionError):
            self._makeOne(highest_visible_tid=2,
                          complete_since_tid=0,
                          data=[(1, 0)])

        # Just right
        c = self._makeOne(highest_visible_tid=3,
                          complete_since_tid=1,
                          data=[(1, 2)])

        self.assertEqual(3, c.highest_visible_tid)
        self.assertEqual(1, c.complete_since_tid)

    def test_complete_to(self):
        # This map has no guarantees about completeness and can
        # have values <= 1.
        old_map = self._makeOne(1, complete_since_tid=None, data=())
        old_map[1] = 1

        # This one does guarantee completeness
        new_map = self._makeOne(4, complete_since_tid=1, data=((1, 2),
                                                               (2, 2)))

        old_map.complete_to(new_map)
        # Check constraints
        old_map.verify(initial=False)
        self.assertEqual(old_map.complete_since_tid, new_map.complete_since_tid)
        self.assertEqual(old_map.highest_visible_tid, new_map.highest_visible_tid)
        self.assertEqual(old_map[1], 2)
        self.assertEqual(old_map[2], 2)

    def test_merge_same_tid(self):
        # A complete map
        old_map = self._makeOne(3, complete_since_tid=2, data=((1, 3),))
        old_map[2] = 1

        # Another one, with more data.
        new_map = self._makeOne(3, complete_since_tid=1, data=((1, 3),
                                                               (3, 2)))

        old_map.merge_same_tid(new_map)
        # Check constraints
        old_map.verify(initial=False)
        self.assertEqual(old_map.complete_since_tid, new_map.complete_since_tid)
        self.assertEqual(old_map.highest_visible_tid, new_map.highest_visible_tid)
        self.assertEqual(old_map[1], 3)
        self.assertEqual(old_map[2], 1)
        self.assertEqual(old_map[3], 2)

    def test_can_hold_old_data(self):
        # Complete maps are only complete over their range:
        # ``complete_since_tid < tid <= highest_visible_tid``.
        # They can have partial incomplete data older than that.

        new_map = self._makeOne(4, complete_since_tid=1, data=((1, 2),
                                                               (2, 2)))
        new_map[3] = 1
        new_map.verify(initial=False)

    def test_merge_older_not_modify_older(self):
        orig = self._makeOne(2, complete_since_tid=None, data=())
        orig[2] = 1
        orig[1] = 1
        old_map = self._makeOne(3, complete_since_tid=2, data=((1, 3),))

        old_map.merge_older_tid(orig)
        self.assertEqual(dict(orig), {1: 1, 2: 1})
        # The merged map has both
        self.assertEqual(dict(old_map), {1: 3, 2: 1})


class TestObjectIndex(TestCase):

    def _makeOne(self, highest_visible_tid=1, data=()):
        return mvcc._ObjectIndex(highest_visible_tid, data=data)

    def test_ctor_empty(self):
        ix = self._makeOne()
        self.assertEqual(1, len(ix.maps))
        self.assertIsInstance(ix.maps[0], mvcc._TransactionRangeObjectIndex)
        self.assertEqual(1, ix.highest_visible_tid)
        self.assertEqual(1, ix.maximum_highest_visible_tid)
        self.assertEqual(1, ix.minimum_highest_visible_tid)

    def test__setitem__out_of_range(self):
        ix = self._makeOne(highest_visible_tid=2)
        ix[1] = 1
        self.assertEqual(ix[1], 1)
        ix[2] = 3
        self.assertNotIn(2, ix)

    def test_ctr_data(self):
        # Too high
        with self.assertRaises(TypeError):
            self._makeOne(highest_visible_tid=2,
                          data=[(1, 3)])

        # Too low
        with self.assertRaises(TypeError):
            # Prior to BTrees 4.7.2, this would be an OverflowError
            # on CPython with C extensions but something else with
            # pure-python mode. 4.7.2 cleaned that up to be TypeError
            # always. Prior to adapting to that change, this hit an assertion
            # error in the pure-python, dict-based case we used on PyPy
            # Now we raise a specific error.
            self._makeOne(highest_visible_tid=2,
                          data=[(1, -1)])

        # Just right
        c = self._makeOne(highest_visible_tid=2,
                          data=[(1, 1)])

        self.assertEqual(2, c.highest_visible_tid)

    def test_polled_changes_first_poll_go_forward(self, new_polled_tid=2):
        initial_tid = 1
        # Initially incomplete.
        ix = self._makeOne(highest_visible_tid=initial_tid)
        initial_map = ix.maps[0]

        # cache some data.
        oid = 1
        oid2 = 2
        oid3 = 3
        ix[oid] = initial_tid
        ix[oid2] = initial_tid

        # Poll for > 1
        complete_since_tid = 1
        new_polled_tid = 2
        changes = [
            (oid, new_polled_tid),
            (oid3, new_polled_tid)
        ]

        ix2 = ix.with_polled_changes(new_polled_tid, complete_since_tid, changes)
        # We did not get the same object back, because the tid changed,
        # and we can't do that.
        self.assertIsNot(ix2, ix)
        del ix
        self.assertIsNot(ix2.maps[0], initial_map)
        # Verifies.
        ix2.verify()
        self.assertEqual(ix2.maximum_highest_visible_tid, new_polled_tid)
        # Keeps connection to oldest map.
        self.assertEqual(ix2.minimum_highest_visible_tid, initial_tid)

        # Incomplete data is preserved, and
        # Complete incoming data is visible
        self.assertEqual(dict(ix2), {
            oid: new_polled_tid,
            oid2: initial_tid,
            oid3: new_polled_tid
        })


    def test_polled_changes_first_poll_get_changes_same_tid(self):
        # Some other entity had a TID in the past (since we began) and
        # they polled and got changes that overlapped our
        # initial set.
        self.test_polled_changes_first_poll_go_forward(new_polled_tid=1)

    def test_polled_changes_does_not_allow_None_changes(self):
        ix = self._makeOne(highest_visible_tid=1)
        with self.assertRaises(AssertionError):
            ix.with_polled_changes(2, 1, None)

    def test_polled_empty_not_complete(self):
        # We assert for this impossible case
        initial_tid = 1
        ix = self._makeOne(highest_visible_tid=1)
        first_poll_tid = 1
        ix = ix.with_polled_changes(highest_visible_tid=first_poll_tid,
                                    complete_since_tid=initial_tid, changes=())
        # We didn't actually change the complete_since value, because without
        # changes we can't actually be sure we're complete.
        self.assertEqual(ix.maps[-1].complete_since_tid, None)

    def test_polled_same_tid_back_complete(self):
        initial_tid = 2

        ix = self._makeOne(highest_visible_tid=initial_tid)

        first_poll_tid = initial_tid + 2

        ix = ix.with_polled_changes(
            highest_visible_tid=first_poll_tid,
            complete_since_tid=initial_tid,
            changes=(
                (1, first_poll_tid),
            ))

        ix2 = ix.with_polled_changes(
            highest_visible_tid=first_poll_tid,
            complete_since_tid=initial_tid - 1,
            changes=(
                (1, first_poll_tid),
                (2, initial_tid)
            )
        )

        self.assertIs(ix2, ix)
        self.assertEqual(ix[1], first_poll_tid)
        self.assertEqual(ix[2], initial_tid)

    def test_polled_newer_tid_after_first(self):
        ix = self._makeOne(highest_visible_tid=1)
        initial_tid = 1
        first_poll_tid = 4

        # Get changes for 3 and above
        ix2 = ix.with_polled_changes(
            highest_visible_tid=first_poll_tid,
            complete_since_tid=initial_tid,
            changes=(
                (1, first_poll_tid),
            ))

        self.assertIsNot(ix, ix2)

        second_poll_tid = 6
        ix2 = ix2.with_polled_changes(
            highest_visible_tid=second_poll_tid,
            complete_since_tid=first_poll_tid,
            changes=(
                (2, second_poll_tid - 1),
                (3, second_poll_tid)
            )
        )

        self.assertEqual(ix2[2], second_poll_tid - 1)
        self.assertEqual(ix2[3], second_poll_tid)
        self.assertEqual(ix2[1], first_poll_tid)



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################


from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest

from .. import interfaces

class TestExceptions(unittest.TestCase):

    def test_DriverNotAvailableError_str(self, klass=interfaces.DriverNotAvailableError):
        from relstorage.adapters.mysql import drivers
        name = 'auto'

        ex = klass(name, drivers, 'reason')

        for s in str(ex), repr(ex):
            self.assertIn(
                "%s: Driver 'auto' is not available (reason=reason). Options: " % (
                    klass.__name__,
                ),
                s
            )
            self.assertIn(
                " 'MySQLdb' (Module: 'MySQLdb'; Available:",
                s
            )

    def test_UnknownDriverError_str(self):
        self.test_DriverNotAvailableError_str(interfaces.UnknownDriverError)

    def test_NoDriversAvailableError_str(self):
        self.test_DriverNotAvailableError_str(interfaces.NoDriversAvailableError)



##############################################################################
#
# Copyright (c) 2004 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from __future__ import absolute_import
from __future__ import print_function

import atexit
import collections
import os
import random
import stat
import struct
import tempfile
import time
import unittest
from hashlib import md5

import transaction
import ZODB.blob
import ZODB.interfaces
import ZODB.tests.StorageTestBase
import ZODB.tests.util
from ZODB.blob import Blob
from ZODB.DB import DB
from ZODB.serialize import referencesf

from relstorage.tests import TestCase
from relstorage.tests.util import USE_SMALL_BLOBS
from relstorage.tests.util import MinimalTestLayer as BaseTestLayer
from relstorage.tests.RecoveryStorage import IteratorDeepCompare
from relstorage.tests.blob.blob_packing import TestBlobPackHistoryPreservingMixin
from relstorage.tests.blob.blob_packing import TestBlobPackHistoryFreeMixin
from relstorage.tests.blob.blob_connection import TestConnectionBlobMixin
from relstorage.tests.blob.blob_importexport import TestBlobImportExportMixin
from relstorage.tests.blob.blob_transaction import TestBlobTransactionMixin
from relstorage.tests.blob.blob_cache import TestBlobCacheMixin
from relstorage.tests.blob.blob_cache import TestBlobCacheExternalCleanupMixin
from relstorage.tests.blob import TestBlobMixin

class BlobTestBase(TestCase,
                   ZODB.tests.StorageTestBase.StorageTestBase):

    def setUp(self):
        super(BlobTestBase, self).setUp()
        try:
            self._storage = self.create_storage() # pylint:disable=no-member
        except:
            self.tearDown()
            raise


class BlobUndoTests(BlobTestBase):

    def testUndoWithoutPreviousVersion(self):
        database = DB(self._storage)
        connection = database.open()
        root = connection.root()
        transaction.begin()
        root['blob'] = Blob()
        transaction.commit()

        database.undo(database.undoLog(0, 1)[0]['id'])
        transaction.commit()
        transaction.manager.manager.free(transaction.get())

        # the blob footprint object should exist no longer
        self.assertRaises(KeyError, root.__getitem__, 'blob')
        database.close()

    def testUndo(self):
        database = DB(self._storage)
        connection = database.open()
        root = connection.root()
        transaction.begin()
        blob = Blob()
        with blob.open('w') as f:
            f.write(b'this is state 1')
        root['blob'] = blob
        transaction.commit()

        transaction.begin()
        blob = root['blob']
        with blob.open('w') as f:
            f.write(b'this is state 2')
        transaction.commit()


        database.undo(database.undoLog(0, 1)[0]['id'])
        transaction.commit()

        with blob.open('r') as f:
            data = f.read()
        self.assertEqual(data, b'this is state 1')

        database.close()

    def testUndoAfterConsumption(self):
        database = DB(self._storage)
        connection = database.open()
        root = connection.root()
        transaction.begin()
        with open('consume1', 'wb') as f: f.write(b'this is state 1')
        blob = Blob()
        blob.consumeFile('consume1')
        root['blob'] = blob
        transaction.commit()

        transaction.begin()
        blob = root['blob']
        with open('consume2', 'wb') as f: f.write(b'this is state 2')
        blob.consumeFile('consume2')
        transaction.commit()

        database.undo(database.undoLog(0, 1)[0]['id'])
        transaction.commit()

        with blob.open('r') as f:
            data = f.read()
        self.assertEqual(data, b'this is state 1')

        database.close()

    def testRedo(self):
        database = DB(self._storage)
        connection = database.open()
        root = connection.root()
        blob = Blob()

        transaction.begin()
        with blob.open('w') as f:
            f.write(b'this is state 1')
        root['blob'] = blob
        transaction.commit()

        transaction.begin()
        blob = root['blob']
        with blob.open('w') as f:
            f.write(b'this is state 2')
        transaction.commit()

        database.undo(database.undoLog(0, 1)[0]['id'])
        transaction.commit()

        with blob.open('r') as f:
            self.assertEqual(f.read(), b'this is state 1')

        database.undo(database.undoLog(0, 1)[0]['id'])
        transaction.commit()

        with blob.open('r') as f:
            self.assertEqual(f.read(), b'this is state 2')

        database.close()

    def testRedoOfCreation(self):
        database = DB(self._storage)
        connection = database.open()
        root = connection.root()
        blob = Blob()

        transaction.begin()
        with blob.open('w') as f:
            f.write(b'this is state 1')
        root['blob'] = blob
        transaction.commit()

        database.undo(database.undoLog(0, 1)[0]['id'])
        transaction.commit()

        self.assertRaises(KeyError, root.__getitem__, 'blob')

        database.undo(database.undoLog(0, 1)[0]['id'])
        transaction.commit()

        with blob.open('r') as f:
            self.assertEqual(f.read(), b'this is state 1')

        database.close()


class RecoveryBlobStorage(BlobTestBase,
                          IteratorDeepCompare):

    _dst = None

    def setUp(self):
        BlobTestBase.setUp(self)
        self._dst = self.create_storage('dest') # pylint:disable=no-member

    def tearDown(self):
        if self._dst is not None:
            self._dst.close()
        BlobTestBase.tearDown(self)

    # Requires a setUp() that creates a self._dst destination storage
    def testSimpleBlobRecovery(self):
        self.assertTrue(
            ZODB.interfaces.IBlobStorageRestoreable.providedBy(
                self._storage)
            )
        db = DB(self._storage)
        conn = db.open()
        root = conn.root()
        root._p_activate()
        root[1] = ZODB.blob.Blob()
        transaction.commit()
        root[2] = ZODB.blob.Blob()
        with root[2].open('w') as f:
            f.write(b'some data')
        transaction.commit()
        root[3] = ZODB.blob.Blob()
        with root[3].open('w') as f:
            f.write(
                (b''.join(struct.pack(">I", random.randint(0, (1<<32)-1))
                          for i in range(random.randint(10000, 20000)))
                )[:-random.randint(1, 4)]
            )
        transaction.commit()
        root[2] = ZODB.blob.Blob()
        with root[2].open('w') as f:
            f.write(b'some other data')
        transaction.commit()
        __traceback_info__ = self._storage, self._dst
        self._dst.copyTransactionsFrom(self._storage)

        self.compare(self._storage, self._dst)

        conn.close()
        db.close()


class LargeBlobTest(BlobTestBase): # pragma: no cover
    """Test large blob upload and download.

    Note that this test excercises the blob storage and only makes sense
    when shared_blob_support=False.

    """
    # Only run when selecting -a 2 or higher, or --all if we're using
    # huge blobs.
    level = 2 if not USE_SMALL_BLOBS else 0
    testsize = 0 # Set on the auto-generated parent class

    with open(__file__, 'rb') as _f:
        # Just use the this module as the source of our data
        # Capture it at import time because test cases may
        # chdir(), and we may not have an absolute path in __file__,
        # depending on how they are run.
        _random_file_data = _f.read().replace(b'\n', b'').split()
    del _f


    def _random_file(self, size, fd):
        """
        Create a random data of at least the given size, writing to fd.

        See
        http://jessenoller.com/2008/05/30/making-re-creatable-random-data-files-really-fast-in-python/
        for the technique used.

        Returns the md5 sum of the file contents for easy comparison.
        """
        def fdata():
            seed = "1092384956781341341234656953214543219"
            words = self._random_file_data
            a = collections.deque(words)
            b = collections.deque(seed)
            while True:
                yield b' '.join(list(a)[0:1024])
                a.rotate(int(b[0]))
                b.rotate(1)
        datagen = fdata()
        bytes = 0
        hasher = md5()
        while bytes < size:
            data = next(datagen)
            hasher.update(data)
            fd.write(data)
            bytes += len(data)
        return hasher.hexdigest()


    def _md5sum(self, fd):
        hasher = md5()
        blocksize = hasher.block_size << 8
        for data in iter(lambda: fd.read(blocksize), b''):
            hasher.update(data)
        return hasher.hexdigest()

    def testLargeBlob(self):
        # Large blobs are chunked into multiple pieces, we want to know
        # if they come out the same way they went in.
        db = DB(self._storage)
        conn = db.open()
        blob = conn.root()[1] = ZODB.blob.Blob()
        blob_file = blob.open('w')
        signature = self._random_file(self.testsize, blob_file)
        blob_file.close()
        transaction.commit()
        conn.close()

        # Clear the cache
        for base, _dir, files in os.walk('.'):
            for f in files:
                if f.endswith('.blob'):
                    ZODB.blob.remove_committed(os.path.join(base, f))

        # Re-download blob
        conn = db.open()
        with conn.root()[1].open('r') as blob:
            self.assertEqual(self._md5sum(blob), signature)

        conn.close()
        db.close()

class TestThingsPreviouslyDocTests(TestBlobMixin,
                                   TestCase):

    def test_packing_with_uncommitted_data_non_undoing(self):
        """
        This covers regression for bug #130459.

        When uncommitted data exists it formerly was written to the root of the
        blob_directory and confused our packing strategy. We now use a separate
        temporary directory that is ignored while packing.
        """

        blob_storage = self.blob_storage
        database = self.database
        connection = database.open()
        root = connection.root()
        root['blob'] = Blob()
        connection.add(root['blob'])
        with root['blob'].open('w') as f: _ = f.write(b'test')

        blob_storage.pack(time.time(), referencesf)

        # Clean up:

        transaction.abort()
        connection.close()
        blob_storage.close()
        database.close()

    def test_packing_with_uncommitted_data_undoing(self):
        """
        This covers regression for bug #130459.

        When uncommitted data exists it formerly was written to the root of the
        blob_directory and confused our packing strategy. We now use a separate
        temporary directory that is ignored while packing.
        """
        blob_storage = self.blob_storage
        database = self.database
        connection = database.open()
        root = connection.root()
        root['blob'] = Blob()
        connection.add(root['blob'])
        with root['blob'].open('w') as f:
            f.write(b'test')

        blob_storage.pack(time.time(), referencesf)

        transaction.abort()
        connection.close()
        database.close()
        blob_storage.close()

    def test_blob_file_permissions(self):
        blob_storage = self.blob_storage
        conn = ZODB.connection(blob_storage)
        conn.root.x = ZODB.blob.Blob(b'test')
        conn.transaction_manager.commit()

        # Blobs have the readability of their parent directories:
        READABLE = stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH
        path = conn.root.x.committed()
        self.assertEqual(
            (os.stat(path).st_mode & READABLE),
            (os.stat(os.path.dirname(path)).st_mode & READABLE))

        # The committed file isn't writable:
        WRITABLE = stat.S_IWUSR | stat.S_IWGRP | stat.S_IWOTH
        self.assertEqual(0, os.stat(path).st_mode & WRITABLE)
        conn.close()

    def test_loadblob_tmpstore(self):
        """
        This is a test for assuring that the TmpStore's loadBlob implementation
        falls back correctly to loadBlob on the backend.
        """

        # First, let's setup a regular database and store a blob:

        blob_storage = self.blob_storage
        database = self.database
        connection = database.open()
        root = connection.root()
        root['blob'] = Blob()
        connection.add(root['blob'])
        with root['blob'].open('w') as f:
            f.write(b'test')
        transaction.commit()
        blob_oid = root['blob']._p_oid
        tid = connection._storage.lastTransaction()

        # Now we open a database with a TmpStore in front:

        from ZODB.Connection import TmpStore
        tmpstore = TmpStore(blob_storage)

        # We can access the blob correctly:
        self.assertEqual(
            tmpstore.loadBlob(blob_oid, tid),
            blob_storage.loadBlob(blob_oid, tid))

        connection.close()
        blob_storage.close()
        tmpstore.close()
        database.close()

    def test_do_not_depend_on_cwd(self):
        bs = self.blob_storage
        here = os.getcwd()
        os.mkdir('evil')
        os.chdir('evil')
        db = DB(bs)
        conn = db.open()
        conn.root()['blob'] = ZODB.blob.Blob()
        with conn.root()['blob'].open('w') as f:
            f.write(b'data')
        transaction.commit()
        os.chdir(here)
        with conn.root()['blob'].open() as f:
            data = f.read()
        self.assertEqual(data, b'data')

        bs.close()

    def test_savepoint_cleanup(self):
        """Make sure savepoint data gets cleaned up."""

        bs = self.blob_storage
        tdir = bs.temporaryDirectory()
        os.listdir(tdir)
        self.assertEmpty(os.listdir(tdir))

        db = self.database
        conn = db.open()
        conn.root().b = ZODB.blob.Blob()
        with conn.root().b.open('w') as f: _ = f.write(b'initial')
        _ = transaction.savepoint()
        self.assertEqual(1, len(os.listdir(tdir)))

        transaction.abort()
        savepoint_dir = os.path.join(tdir, 'savepoint')
        self.assertFalse(
            os.path.exists(savepoint_dir) and len(os.listdir(savepoint_dir)) > 0)

        conn.root().b = ZODB.blob.Blob()
        with conn.root().b.open('w') as f: _ = f.write(b'initial')
        transaction.commit()
        with conn.root().b.open('w') as f: _ = f.write(b'1')
        _ = transaction.savepoint()
        transaction.abort()
        self.assertFalse(os.path.exists(savepoint_dir) and len(os.listdir(savepoint_dir)) > 0)

        db.close()


class BlobTestLayer(BaseTestLayer):

    __bases__ = ()
    __module__ = ''

    def __init__(self, name):
        BaseTestLayer.__init__(self, name)

    def setUp(self):
        self.here = os.getcwd()
        # In the past, this used dir=os.getcwd(), meaning we
        # could pollute our source directory if cleanup failed.
        # Why?
        self.tmp = tempfile.mkdtemp(self.__name__)
        os.chdir(self.tmp)

        # sigh. tearDown isn't called when a layer is run in a sub-process.
        atexit.register(clean, self.tmp)

    def tearDown(self):
        os.chdir(self.here)
        rmtree(self.tmp)

    def testSetUp(self):
        transaction.abort()

    def testTearDown(self):
        transaction.abort()


def clean(tmp):
    if os.path.isdir(tmp):
        rmtree(tmp)

def rmtree(path):
    """Remove a tree without causing Windows file access errors"""
    # copied from setupstack.py
    for cpath, dirs, files in os.walk(path, False):
        for fname in files:
            fname = os.path.join(cpath, fname)
            os.chmod(fname, stat.S_IWUSR)
            os.remove(fname)
        for dname in dirs:
            dname = os.path.join(cpath, dname)
            os.rmdir(dname)
    os.rmdir(path)


class TestBlobPackingHP(TestBlobPackHistoryPreservingMixin,
                        TestCase):
    pass

class TestBlobPackingHF(TestBlobPackHistoryFreeMixin,
                        TestCase):
    pass

class TestConnectionBlob(TestConnectionBlobMixin,
                         TestCase):
    pass

class TestBlobImportExport(TestBlobImportExportMixin,
                           TestCase):
    pass

class TestBlobTransaction(TestBlobTransactionMixin,
                          TestCase):
    pass

class TestBlobCache(TestBlobCacheMixin,
                    TestCase):
    pass

class TestBlobCacheExternalCleanup(TestBlobCacheExternalCleanupMixin,
                                   TestCase):
    pass

def storage_reusable_suite(prefix, factory,
                           test_blob_storage_recovery=False,
                           test_packing=False,
                           test_undo=True,
                           keep_history=True,
                           test_blob_cache=False,
                           large_blob_size=None,
                           storage_is_available=True):
    """Return a test suite for a generic IBlobStorage.

    Pass a factory taking a name and a blob directory name.
    """

    suite = unittest.TestSuite()

    def create_storage(name='data', blob_dir=None, zap=True, **kw):
        if blob_dir is None:
            blob_dir = '%s.blobs' % name
        storage = factory(name, blob_dir, **kw)
        if zap:
            storage.zap_all()
        return storage

    def add_test_based_on_test_class(klass, **attr):
        attr.update(create_storage=staticmethod(create_storage))
        new_class = klass.__class__(
            prefix + klass.__name__,
            (klass, ),
            attr,
        )
        new_class.__module__ = klass.__module__
        new_class = unittest.skipUnless(storage_is_available, str(storage_is_available))(new_class)
        suite.addTest(unittest.makeSuite(new_class))

    add_test_based_on_test_class(TestBlobTransaction)
    add_test_based_on_test_class(TestBlobImportExport)
    add_test_based_on_test_class(TestThingsPreviouslyDocTests)
    add_test_based_on_test_class(TestConnectionBlob)
    if test_blob_cache:
        add_test_based_on_test_class(TestBlobCache)
        add_test_based_on_test_class(TestBlobCacheExternalCleanup)
    if test_packing:
        base = TestBlobPackingHP if keep_history else TestBlobPackingHF
        add_test_based_on_test_class(base)
    if test_blob_storage_recovery:
        add_test_based_on_test_class(RecoveryBlobStorage)
    if test_undo:
        add_test_based_on_test_class(BlobUndoTests)
    if large_blob_size:
        add_test_based_on_test_class(LargeBlobTest, testsize=large_blob_size)

    suite.layer = BlobTestLayer('BlobTests')

    return suite



import logging
import sys
import time

import transaction
from BTrees.OOBTree import OOBTree  # pylint:disable=no-name-in-module,import-error
from ZODB.DB import DB

from relstorage.adapters.postgresql import PostgreSQLAdapter
from relstorage.storage import RelStorage

log = logging.getLogger('bigpack')


def fill_db(db):
    conn = db.open()
    root = conn.root()
    root['tree'] = tree = OOBTree()
    for j in range(1, 101):
        log.info("Filling %d/100", j)
        for i in range(0, 100000, j):
            tree[i] = []
        transaction.commit()


def main():
    logging.basicConfig(
        stream=sys.stderr,
        level=logging.DEBUG,
        format='%(asctime)s [%(name)s] %(levelname)s %(message)s')
    log.info("Opening")
    adapter = PostgreSQLAdapter()
    storage = RelStorage(adapter)
    db = DB(storage)
    log.info("Filling")
    fill_db(db)
    log.info("Packing")
    start = time.time()
    db.pack()
    end = time.time()
    log.info("Packed in %0.3f seconds", end - start)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from unittest import skipUnless

from relstorage.adapters.tests import test_drivers
from relstorage.tests import TestCase

from ..drivers import Sqlite3GeventDriver

class Driver(Sqlite3GeventDriver):
    def get_driver_module(self):
        import sqlite3
        return sqlite3

    def connect(self): # pylint:disable=arguments-differ
        return self._connect(':memory:')

@skipUnless(Driver.STATIC_AVAILABLE, "Driver not available")
class TestSqlite3GeventDriver(test_drivers.IDBDriverSupportsCriticalTestMixin,
                              TestCase):

    def _makeOne(self):
        return Driver()



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Database schema installers
"""
import abc
import os
import glob
import sys

from functools import partial
from hashlib import md5

from ZODB.POSException import StorageError

from .._compat import ABC
from ._util import DatabaseHelpersMixin
from ._util import query_property
from ._util import noop_when_history_free

from .connmanager import connection_callback

from .sql import View
from .sql import Table
from .sql import TemporaryTable
from .sql import Column
from .sql import HistoryVariantTable
from .sql import OID
from .sql import TID
from .sql import State
from .sql import Boolean
from .sql import BinaryString
from .sql import Char
from .sql import ColumnExpression
from .sql import it

StateSize = OID

logger = __import__('logging').getLogger(__name__)

tmpl_property = partial(query_property,
                        property_suffix='_TMPLS',
                        lazy_suffix='_TMPL')

class Schema(object):
    current_object = Table(
        'current_object',
        Column('zoid', OID),
        Column('tid', TID)
    )

    object_state = Table(
        'object_state',
        Column('zoid', OID),
        Column('tid', TID),
        Column('state', State),
        Column('state_size', StateSize),
        # These two only exist in history preserving.
        # TODO: Wrap this in a HistoryVariantTable
        Column('prev_tid', TID),
        Column('md5'),
    )

    blob_chunk = Table(
        'blob_chunk',
        Column('zoid'),
        Column('tid'),
        Column('chunk_num'),
        Column('chunk')
    )

    # Does the right thing whether history free or preserving
    all_current_object = HistoryVariantTable(
        current_object,
        object_state,
    )

    current_object_state = View(
        'current_object_state',
        object_state.inner_join(
            current_object
        ).using(
            current_object.c.zoid,
            current_object.c.tid
        ).select(
            object_state.c.zoid,
            object_state.c.tid,
            it.c.state
        )
    )

    # Does the right thing whether history free or preserving
    all_current_object_state = HistoryVariantTable(
        current_object_state,
        object_state
    )

    temp_store = TemporaryTable(
        'temp_store',
        Column('zoid', OID, primary_key=True),
        Column('prev_tid', TID, nullable=False),
        Column('md5', Char(32)), # Only used when keep_history=True
        Column('state', State)
    )

    temp_blob_chunk = TemporaryTable(
        'temp_blob_chunk',
        Column('zoid'),
        Column('chunk_num'),
        Column('chunk'),
    )

    transaction = Table(
        'transaction',
        Column('tid', TID, primary_key=True),
        Column('packed', Boolean, nullable=False, default=False),
        Column('is_empty', Boolean, nullable=False, default=False),
        Column('username', BinaryString, nullable=False),
        Column('description', BinaryString, nullable=False),
        Column('extension', BinaryString),
    )

    commit_row_lock = Table(
        'commit_row_lock',
        Column('tid', TID, primary_key=True),
    )

    all_transaction = HistoryVariantTable(
        transaction,
        object_state,
    )

    # Not all databases will need this table
    new_oid = Table(
        'new_oid',
        Column('zoid', OID, primary_key=True, auto_increment=True)
    )

    # Packing
    object_ref = Table(
        'object_ref',
        Column('tid', TID),
    )

    object_refs_added = Table(
        'object_refs_added',
        Column('tid', TID)
    )

    pack_object = Table(
        'pack_object',
        Column('zoid', OID, primary_key=True),
        Column('keep', Boolean, nullable=False),
        Column('keep_tid', TID, nullable=False),
        Column('visited', Boolean, nullable=False, default=False)
    )


class AbstractSchemaInstaller(DatabaseHelpersMixin,
                              ABC):

    # Names of schema objects: Wherever we write them, they should be
    # in lower case, even if the database internally changes the case.
    # (In general, they should be written to match the output of
    # :meth:`_normalize_schema_object_names`.)

    # Keep this list in the same order as the schema scripts,
    # for dependency (Foreign Key) purposes.
    # These must be lower case, and all queries we write must
    # use lower case table names (MySQL has weird casing rules,
    # but this is also how we do comparisons in Python on table metadata.)
    all_tables = (
        # History-free row lock table for commits.
        # The alternative is to also create the transaction table
        # in HF mode.
        'commit_row_lock',
        'pack_lock',
        'transaction',
        'new_oid', # If you use a sequence, remove this
        'object_state',
        'blob_chunk',
        'current_object',
        'object_ref',
        'object_refs_added',
        'pack_object',
        'pack_state',
        'pack_state_tid',
        'temp_store',
        'temp_blob_chunk',
        'temp_pack_visit',
        'temp_undo',
    )

    all_sequences = ()

    all_views = (
        'current_object_state',
    )

    # Tables that might exist, but which are unused and obsolete.
    # These can/should be dropped, and names shouldn't be reused.
    obsolete_tables = (
        'commit_lock',
    )

    database_type = None  # provided by a subclass

    # A dictionary from procedure name to procedure definition.
    # We populate this automatically at construction time based on
    # our history-keeping state.
    procedures = None

    # Class variable.
    # {keep_history: {proc_name: proc_source}}
    # subclasses *must* define if they want caching.
    _PROCEDURES = None # type: dict

    def __init__(self, connmanager, runner, keep_history):
        self.connmanager = connmanager
        self.driver = connmanager.driver
        self.keep_history = keep_history
        self.runner = runner.with_format_vars(
            tid_type=self.COLTYPE_OID_TID,
            oid_type=self.COLTYPE_OID_TID,
            binary_string_type=self.COLTYPE_BINARY_STRING,
            blob_chunk_type=self.COLTYPE_BLOB_CHUNK,
            blob_chunk_num_type=self.COLTYPE_BLOB_CHUNK_NUM,
            md5_type=self.COLTYPE_MD5,
            state_type=self.COLTYPE_STATE,
            state_size_type=self.COLTYPE_STATE_SIZE,
            transactional_suffix=self.TRANSACTIONAL_TABLE_SUFFIX,
        )

        # if subclasses don't define, we don't cache.
        proc_cache = self._PROCEDURES if self._PROCEDURES is not None else {}
        if self.keep_history not in proc_cache:
            proc_cache[self.keep_history] = self._read_proc_files()
        self.procedures = proc_cache[self.keep_history]

    def _read_proc_files(self):
        """
        Read the procedure files appropriate for the *keep_history*
        setting and return a dictionary from procedure name to
        procedure definition source.

        The files come from the ``procs`` subdirectory beside the
        ``__file__`` in which the class of *self* is defined, and
        either the ``hf`` or ``hp`` folders within ``procs``
        (depending on ``self.keep_history``). The file name, minus the
        ``.sql`` extension, must match the name of the procedure. All
        ``.sql`` files in the ``procs`` directory are used for both
        history states, and then files in the ``hf`` or ``hp``
        directory are added to that; the file names in ``procs`` and ``h[pf]``
        must not be duplicates. Further, the file name (minus extension)
        must appear in the source.

        The file source is read in text mode and stripped, but other than that
        is unprocessed.
        """

        # TODO: importlib.resources or its backport importlib_resources

        self_dir = os.path.dirname(sys.modules[type(self).__module__].__file__)

        generic_proc_dir = os.path.join(self_dir, 'procs')
        specific_proc_dir = os.path.join(generic_proc_dir, 'hp' if self.keep_history else 'hf')
        logger.debug(
            "Reading stored procedures from %s and %s",
            generic_proc_dir, specific_proc_dir
        )
        proc_files = []
        for d in generic_proc_dir, specific_proc_dir:
            proc_files.extend(
                glob.glob(os.path.join(d, "*.sql"))
            )

        # Make sure there's no dups, that's probably an accident
        assert len(proc_files) == len(set(proc_files))

        procedures = {}
        for proc_file_name in proc_files:
            with open(proc_file_name, "rt") as f:
                source = f.read().strip()
            proc_name = os.path.splitext(os.path.basename(proc_file_name))[0]
            __traceback_info__ = proc_file_name, proc_name
            assert proc_name in source
            procedures[proc_name] = source

        return procedures

    @staticmethod
    def _checksum_for_str(stmt):
        return md5(
            stmt.encode('ascii')
            if not isinstance(stmt, bytes)
            else stmt
        ).hexdigest()

    @abc.abstractmethod
    def list_tables(self, cursor):
        raise NotImplementedError

    @abc.abstractmethod
    def list_sequences(self, cursor):
        raise NotImplementedError

    @abc.abstractmethod
    def list_procedures(self, cursor):
        raise NotImplementedError

    @abc.abstractmethod
    def list_views(self, cursor):
        return ()

    @abc.abstractmethod
    def get_database_name(self, cursor):
        raise NotImplementedError

    _create_commit_row_lock_query = Schema.commit_row_lock.create()

    def _create_commit_row_lock(self, cursor):
        """
        Create the global lock held during commit.

        (MySQL and PostgreSQL do this differently.)
        """
        self._create_commit_row_lock_query.execute(cursor)

    @abc.abstractmethod
    def _create_pack_lock(self, cursor):
        """
        Create the global lock held during pack.

        (MySQL and PostgreSQL do this differently.)
        """
        raise NotImplementedError()

    # TODO: Finish porting the CREATE statements to use the table
    # definitions in Schema and away from string templates.

    #: The type of the column used to hold transaction IDs
    #: and object IDs (64-bit integers).
    COLTYPE_OID_TID = 'BIGINT'
    #: The type of the column used to hold binary strings.
    #: Our default is appropriate for PostgreSQL.
    COLTYPE_BINARY_STRING = 'BYTEA'
    COLTYPE_STATE = COLTYPE_BINARY_STRING
    COLTYPE_STATE_SIZE = 'BIGINT'
    #: The type of the column used to number blob chunks.
    COLTYPE_BLOB_CHUNK_NUM = 'BIGINT'
    #: The type of the column used to store blob chunks.
    COLTYPE_BLOB_CHUNK = 'OID'
    #: The type of the column used to store MD5 hash strings.
    COLTYPE_MD5 = 'CHAR(32)'
    #: The suffix needed (after the closing ')') to make sure a
    #: table behaves in a transactional manner.
    #: Our default is appropriate for PostgreSQL.
    TRANSACTIONAL_TABLE_SUFFIX = ''

    _create_transaction_query = Schema.transaction.create()

    @noop_when_history_free
    def _create_transaction(self, cursor):
        """
        The transaction table lists all the transactions in the database.

        This table is only used for history-preserving databases.
        """
        self._create_transaction_query.execute(cursor)

    # NOTE: Prior to MySQL 8.0.16, CHECK constraints
    # are ignored at creation time and dropped. Thus if you upgrade
    # an existing 5.7 schema to 8, constraints will not be enforced,
    # but if you create a new schema under 8, these constraints will
    # be enforced.

    CREATE_OBJECT_STATE_TMPLS = (
        """
        CREATE TABLE object_state (
            zoid        {oid_type} NOT NULL,
            tid         {tid_type} NOT NULL
                           REFERENCES "transaction",
            prev_tid    {tid_type} NOT NULL
                           REFERENCES "transaction",
            md5         {md5_type},
            state_size  {state_size_type} NOT NULL,
            state       {state_type},
            CONSTRAINT object_state_pk
                PRIMARY KEY (zoid, tid),
            CHECK (tid > 0),
            CHECK (state_size >= 0)
        ) {transactional_suffix};
        CREATE INDEX object_state_tid ON object_state (tid);
        CREATE INDEX object_state_prev_tid ON object_state (prev_tid);
        """,
        """
        CREATE TABLE object_state (
            zoid        {oid_type} NOT NULL PRIMARY KEY,
            tid         {tid_type} NOT NULL,
            state_size  {state_size_type} NOT NULL,
            state       {state_type},
            CHECK (tid > 0),
            CHECK (state_size >= 0)
        ) {transactional_suffix};
        CREATE INDEX object_state_tid ON object_state (tid);
        """
    )

    CREATE_OBJECT_STATE_TMPL = tmpl_property('CREATE_OBJECT_STATE')

    def _create_object_state(self, cursor):
        """
        Create the table holding all object states for all transactions.

        If the schema is history-free, only store the current state.
        History-preserving schemas may have a NULL `object_state` to represent
        uncreation.
        """
        self.runner.run_script(cursor, self.CREATE_OBJECT_STATE_TMPL)

    CREATE_BLOB_CHUNK_TMPLS = (
        """
        CREATE TABLE blob_chunk (
            zoid        {oid_type} NOT NULL,
            tid         {tid_type} NOT NULL,
            chunk_num   {blob_chunk_num_type} NOT NULL,
            chunk       {blob_chunk_type} NOT NULL,
            CONSTRAINT blob_chunk_pk
                PRIMARY KEY (zoid, tid, chunk_num),
            CONSTRAINT blob_chunk_fk
                FOREIGN KEY (zoid, tid)
                REFERENCES object_state (zoid, tid)
                ON DELETE CASCADE
        ) {transactional_suffix};
        CREATE INDEX blob_chunk_lookup ON blob_chunk (zoid, tid);
        """,
        """
        CREATE TABLE blob_chunk (
            zoid        {oid_type} NOT NULL,
            chunk_num   {blob_chunk_num_type} NOT NULL,
            tid         {tid_type} NOT NULL,
            chunk       {blob_chunk_type} NOT NULL,
            CONSTRAINT blob_chunk_pk
                PRIMARY KEY (zoid, chunk_num),
            CONSTRAINT  blob_chunk_fk
                FOREIGN KEY (zoid)
                REFERENCES object_state (zoid)
                ON DELETE CASCADE
        ) {transactional_suffix};
        CREATE INDEX blob_chunk_lookup ON blob_chunk (zoid);
        """
    )

    CREATE_BLOB_CHUNK_TMPL = tmpl_property('CREATE_BLOB_CHUNK')

    def _create_blob_chunk(self, cursor):
        """
        Create the table holding all blob states for all transactions.

        If the schema is history-free, only store the current state.
        """
        self.runner.run_script(cursor, self.CREATE_BLOB_CHUNK_TMPL)

    CREATE_CURRENT_OBJECT_TMPL = """
    CREATE TABLE current_object (
        zoid        {oid_type} NOT NULL PRIMARY KEY,
        tid         {tid_type} NOT NULL,
        FOREIGN KEY (zoid, tid)
            REFERENCES object_state (zoid, tid)
    ) {transactional_suffix};
    CREATE INDEX current_object_tid ON current_object (tid);
    """

    @noop_when_history_free
    def _create_current_object(self, cursor):
        """
        Table that stores pointers to the current object state.

        This table is only used for history-preserving databases.
        """
        self.runner.run_script(cursor, self.CREATE_CURRENT_OBJECT_TMPL)

    CREATE_OBJECT_REF_TMPLS = (
        """
        CREATE TABLE object_ref (
            zoid        {oid_type} NOT NULL,
            tid         {tid_type} NOT NULL,
            to_zoid     {oid_type} NOT NULL,
            PRIMARY KEY (tid, zoid, to_zoid)
        ) {transactional_suffix};
        """,
        """
        CREATE TABLE object_ref (
            zoid        {oid_type} NOT NULL,
            to_zoid     {tid_type} NOT NULL,
            tid         {oid_type} NOT NULL,
            PRIMARY KEY (zoid, to_zoid)
        ) {transactional_suffix};
        """
    )

    CREATE_OBJECT_REF_TMPL = tmpl_property('CREATE_OBJECT_REF')

    def _create_object_ref(self, cursor):
        """
        A list of referenced OIDs from each object_state. This
        table is populated as needed during packing. To prevent unnecessary
        table locking, it does not use foreign keys, which is safe because
        rows in object_state are never modified once committed, and rows are
        removed from object_state only by packing.
        """
        self.runner.run_script(cursor, self.CREATE_OBJECT_REF_TMPL)


    CREATE_OBJECT_REFS_ADDED_TMPLS = (
        """
        CREATE TABLE object_refs_added (
            tid         {tid_type} NOT NULL PRIMARY KEY
        ) {transactional_suffix};
        """,
        """
        CREATE TABLE object_refs_added (
            zoid        {oid_type} NOT NULL PRIMARY KEY,
            tid         {tid_type} NOT NULL
        ) {transactional_suffix}
        """
    )

    CREATE_OBJECT_REFS_ADDED_TMPL = tmpl_property('CREATE_OBJECT_REFS_ADDED')

    def _create_object_refs_added(self, cursor):
        """
        The object_refs_added table tracks whether object_refs has been
        populated for all states in a given transaction. An entry is added
        only when the work is finished. To prevent unnecessary table locking,
        it does not use foreign keys, which is safe because object states are
        never added to a transaction once committed, and rows are removed
        from the transaction table only by packing.
        """
        self.runner.run_script(cursor, self.CREATE_OBJECT_REFS_ADDED_TMPL)


    CREATE_PACK_OBJECT_IX_TMPL = """
    CREATE INDEX pack_object_keep_zoid ON pack_object (keep, zoid)
    """

    _create_pack_object_query = Schema.pack_object.create()

    def _create_pack_object(self, cursor):
        """
        pack_object contains temporary state during garbage collection: The
        list of all objects, a flag signifying whether the object should be
        kept, and a flag signifying whether the object's references have been
        visited. The keep_tid field specifies the current revision of the
        object.
        """
        self._create_pack_object_query.execute(cursor)
        # Some databases (pg) have multiple statements
        self.runner.run_script(cursor, self.CREATE_PACK_OBJECT_IX_TMPL)

    CREATE_PACK_STATE_TMPL = """
    CREATE TABLE pack_state (
        tid         {tid_type} NOT NULL,
        zoid        {oid_type} NOT NULL,
        PRIMARY KEY (tid, zoid)
    ) {transactional_suffix};
    """

    @noop_when_history_free
    def _create_pack_state(self, cursor):
        """
        Temporary state populated during pre-packing.

        This is only used in history-preserving databases.

        This table is poorly named. What it actually holds is the set
        of objects, along with their maximum TID, that are potentially
        eligible to be discarded because their most recent change
        (maximum TID) is earlier than the pack time.
        """
        self.runner.run_script(cursor, self.CREATE_PACK_STATE_TMPL)

    CREATE_PACK_STATE_TID_TMPL = """
    CREATE TABLE pack_state_tid (
        tid {tid_type} NOT NULL PRIMARY KEY
    ) {transactional_suffix};
    """

    @noop_when_history_free
    def _create_pack_state_tid(self, cursor):
        """
        Temporary state during pre-packing:

        This is only used in history-preserving databases.

        This table is poorly named. What it actually holds is simply a
        summary of the distinct transaction IDs found in
        ``pack_state``. In other words, it's the list of transaction
        IDs that are eligible to be discarded.
        """
        self.runner.run_script(cursor, self.CREATE_PACK_STATE_TID_TMPL)


    # Most databases handle temp tables on a session-by-session
    # basis.
    def _create_temp_store(self, _cursor):
        """States that will soon be stored."""
        return

    def _create_temp_blob_chunk(self, _cursor):
        """
        Temporary state during packing: a list of objects
        whose references need to be examined.
        """
        return

    def _create_temp_pack_visit(self, _cursor):
        return

    def _create_temp_undo(self, _cursor):
        """
        Temporary state during undo: a list of objects
        to be undone and the tid of the undone state.
        """
        return

    _create_current_object_state_query = Schema.current_object_state.create()

    @noop_when_history_free
    def _create_current_object_state(self, cursor):
        self._create_current_object_state_query.execute(cursor)

    def _init_after_create(self, cursor):
        """
        Create a special '0' transaction to represent object creation. The
        '0' transaction is often referenced by object_state.prev_tid, but
        never by object_state.tid. (Only in history-preserving databases.)

        In all databases, populates the ``commit_row_lock`` table
        with a single row to use as a global lock at commit time.
        """
        if self.keep_history:
            stmt = Schema.transaction.insert(
                Schema.transaction.c.tid,
                Schema.transaction.c.username,
                Schema.transaction.c.description
            ).bind(self).compiled()
            stmt.execute(
                cursor,
                (0,
                 self.driver.Binary(b'system'),
                 self.driver.Binary(b'special transaction for object creation'))
            )

        stmt = """
        INSERT INTO commit_row_lock (tid)
        VALUES (0);
        """
        self.runner.run_script(cursor, stmt)

    @abc.abstractmethod
    def _reset_oid(self, cursor):
        raise NotImplementedError()

    def _normalize_schema_object_names(self, names):
        return [n.lower() for n in names]

    def _create_schema_objects(self, cursor, all_object_names, existing_object_names):
        # Normalize names
        existing_object_names = set(self._normalize_schema_object_names(existing_object_names))
        todo = set(all_object_names) - existing_object_names
        # Preserve order
        todo = [t for t in all_object_names if t in todo]
        __traceback_info__ = todo, existing_object_names, all_object_names
        for table in todo:
            meth = getattr(self, '_create_' + table)
            meth(cursor)
        return todo, existing_object_names

    def create_tables(self, cursor, existing_tables=()):
        """Create the database tables."""
        _, existing_tables = self._create_schema_objects(cursor, self.all_tables, existing_tables)

        if self.keep_history and 'transaction' not in existing_tables:
            self._init_after_create(cursor)
        elif not self.keep_history and 'commit_row_lock' not in existing_tables:
            self._init_after_create(cursor)

        tables = self.list_tables(cursor)
        self.check_compatibility(cursor, tables)
        return tables

    def create_procedures(self, cursor):
        "Subclasses should override"

    def create_triggers(self, cursor):
        "Subclasses should override"

    def create_sequences(self, cursor):
        "Subclasses should override."
        self._create_schema_objects(cursor,
                                    self.all_sequences,
                                    self.list_sequences(cursor))

    def create_views(self, cursor):
        self._create_schema_objects(cursor,
                                    self.all_views,
                                    self.list_views(cursor))

    @connection_callback(read_only=False, application_name="RS: PrepareSchema")
    def _prepare_with_connection(self, conn, cursor): # pylint:disable=unused-argument
        # XXX: We can generalize this to handle triggers, procs, etc,
        # to make subclasses have easier time.
        existing_tables = self.list_tables(cursor)
        __traceback_info__ = existing_tables
        all_tables = self.create_tables(cursor, existing_tables)
        __traceback_info__ = existing_tables, all_tables
        if 'transaction' in self._normalize_schema_object_names(existing_tables):
            self.update_schema(cursor, existing_tables)

        self.create_sequences(cursor)
        self.create_procedures(cursor)
        self.create_triggers(cursor)
        self.create_views(cursor)

    def prepare(self):
        self.connmanager.open_and_call(self._prepare_with_connection)

    def verify(self):
        self.connmanager.open_and_call(self._verify)

    @connection_callback(read_only=True, application_name='RS: VerifySchema')
    def _verify(self, conn, cursor): # pylint:disable=unused-argument
        tables = self.list_tables(cursor)
        self.check_compatibility(cursor, tables)

    def check_compatibility(self, cursor, tables): # pylint:disable=unused-argument
        tables = self._normalize_schema_object_names(tables)
        if self.keep_history:
            if 'transaction' not in tables and 'current_object' not in tables:
                raise StorageError(
                    "Schema mismatch: a history-preserving adapter "
                    "can not connect to a history-free database. "
                    "If you need to convert, use the zodbconvert utility."
                )
        else:
            if 'transaction' in tables and 'current_object' in tables:
                raise StorageError(
                    "Schema mismatch: a history-free adapter "
                    "can not connect to a history-preserving database. "
                    "If you need to convert, use the zodbconvert utility."
                )
        if 'blob_chunk' not in tables:
            raise StorageError(
                "Schema mismatch; please create the blob_chunk tables. "
                "See migration instructions for RelStorage 1.5. "
                "All tables: %s" % (tables,)
            )

    def update_schema(self, cursor, tables): # pylint:disable=unused-argument
        """
        Perform any migration steps that are needed to make a schema
        that has already been created some time in the past match
        what would currently be installed.

        Subclasses may override.
        """

        # Currently we only take care of renaming the `transaction.empty`
        # column (from RelStorage 2.x and earlier) to `transaction.is_empty`
        # as used in RelStorage 3.x.
        if self._needs_transaction_empty_update(cursor):
            cursor.execute(self._rename_transaction_empty_stmt)

    _blank_transaction_query = Schema.transaction.select(
        ColumnExpression('*')
    ).where(Schema.transaction.c.tid < 0)

    def _needs_transaction_empty_update(self, cursor):
        # Get a description of the table, but don't actually return
        # any rows.
        if not self.keep_history:
            return False

        self._blank_transaction_query.execute(cursor)
        columns = self._column_descriptions(cursor)
        # Make sure to read the (empty) result, some drivers (CMySQLConnector)
        # are picky about that and won't let you close a cursor without reading
        # everything.
        cursor.fetchall()

        for column_descr in columns:
            if column_descr.name.lower() == 'is_empty':
                # Yay, nothing to do.
                return False

        # If we get here, the is_empty column isn't present.
        # Must rename it.
        return True

    _rename_transaction_empty_stmt = 'ALTER TABLE transaction RENAME COLUMN empty TO is_empty'

    # Subclasses can redefine these.
    _slow_zap_all_tbl_stmt = _zap_all_tbl_stmt = 'DELETE FROM "%s"'


    def zap_all(self, reset_oid=True, slow=False):
        """
        Clear all data out of the database.

        :keyword bool slow: If True (*not* the default) then database
            specific optimizations will be skipped and rows will simply be
            DELETEd. This is helpful when other connections might be open and
            holding some kind of locks.
        """
        stmt = self._zap_all_tbl_stmt if not slow else self._slow_zap_all_tbl_stmt

        def zap_all(_conn, cursor):
            existent = set(self.list_tables(cursor))
            to_zap = {} # {normalized_name: recorded_name}
            self._before_zap_all_tables(cursor, existent, slow)
            for possible_table in existent:
                norm_table = self._normalize_schema_object_names([possible_table])[0]
                if norm_table.startswith('temp_'):
                    continue
                if norm_table in self.all_tables:
                    to_zap[norm_table] = possible_table

            # Do in reverse order because that's how Foreign Key
            # constraints are set up.
            for possible_table in reversed(self.all_tables):
                if possible_table in to_zap:
                    table_stmt = stmt % to_zap[possible_table]
                    logger.debug(table_stmt)
                    __traceback_info__ = table_stmt
                    cursor.execute(table_stmt)
            logger.debug("Done deleting from tables.")

            self._after_zap_all_tables(cursor, slow)

            if reset_oid:
                logger.debug("Running OID reset script.")
                self._reset_oid(cursor)
                logger.debug("Done running OID reset script.")

        self.connmanager.open_and_call(zap_all)

    # Hooks for subclasses

    def _before_zap_all_tables(self, cursor, tables, slow=False):
        logger.debug("Before zapping existing tables (%s) with %s; slow: %s",
                     tables, cursor, slow)

    def _after_zap_all_tables(self, cursor, slow=False):
        logger.debug("Running init script. Slow: %s", slow)
        self._init_after_create(cursor)
        logger.debug("Done running init script.")

    DROP_TABLE_TMPL = 'DROP TABLE "{table}"'

    def _drop_all(self, _conn, cursor):
        for view in self.list_views(cursor):
            if view in self.all_views:
                cursor.execute('DROP VIEW %s' % view)

        existent = set(self.list_tables(cursor))
        todo = list(self.all_tables)
        todo.reverse()
        for table in todo:
            if table in existent:
                cursor.execute(self.DROP_TABLE_TMPL.format(table=table))
        for sequence in self.list_sequences(cursor):
            cursor.execute("DROP SEQUENCE %s" % sequence)

    def drop_all(self):
        """Drop all tables and sequences."""
        self.connmanager.open_and_call(self._drop_all)



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Transaction iterator support.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from ZODB.BaseStorage import DataRecord
from ZODB.BaseStorage import TransactionRecord
from ZODB.interfaces import StorageStopIteration
from ZODB.utils import p64 as int64_to_8bytes
from ZODB.utils import u64 as bytes8_to_int64

from relstorage._compat import loads
from relstorage.adapters.connections import LoadConnection

logger = __import__('logging').getLogger(__name__)


class _TransactionIterator(object):
    """
    Iterate over the transactions in a RelStorage instance.
    """

    __slots__ = (
        '_adapter',
        '_cursor',
        '_closed',
        '_transactions',
        '_index',
    )

    def __init__(self, adapter, load_connection, start, stop):
        self._adapter = adapter
        self._cursor = load_connection.cursor
        self._closed = False

        if start is not None:
            start_int = bytes8_to_int64(start)
        else:
            start_int = 1
        if stop is not None:
            stop_int = bytes8_to_int64(stop)
        else:
            stop_int = None

        # _transactions: [(tid, username, description, extension, packed)]
        with load_connection.server_side_cursor() as cursor:
            self._transactions = adapter.dbiter.iter_transactions_range(
                cursor, start_int, stop_int)
        self._index = 0

    def close(self):
        self._closed = True
        self._cursor = None
        self._transactions = ()

    def __del__(self):
        # belt-and-suspenders, effective on CPython
        # TODO: Issue a warning if we get here without being closed.
        self.close()

    def iterator(self):
        return self

    def __iter__(self):
        return self

    def __len__(self):
        return len(self._transactions)

    def __getitem__(self, n):
        self._index = n
        return next(self)

    def next(self):
        if self._index >= len(self._transactions):
            self.close() # Don't leak our connection
            raise StorageStopIteration()
        if self._closed:
            raise IOError("TransactionIterator already closed")
        params = self._transactions[self._index]
        res = RelStorageTransactionRecord(
            self,
            params.tid_int,
            params.username,
            params.description,
            params.extension,
            params.packed
        )
        self._index += 1
        return res

    __next__ = next


class HistoryPreservingTransactionIterator(_TransactionIterator):
    """
    Opens a distinct load connection to read transactions.

    Closes this connection when it is closed.
    """
    __slots__ = (
        '_conn',
    )

    def __init__(self, adapter, start, stop):
        self._conn = load_connection = LoadConnection(adapter.connmanager)
        super(HistoryPreservingTransactionIterator, self).__init__(
            adapter, load_connection, start, stop)

    def close(self):
        try:
            if self._conn is not None:
                self._conn.drop()
        finally:
            self._conn = None
            super(HistoryPreservingTransactionIterator, self).close()


class HistoryFreeTransactionIterator(_TransactionIterator):
    """
    Uses the given load connection cursor. Does not close the
    connection or cursor. This way multiple iterators can be
    consistent with each other.

    If we do not do this, then multiple iterators may be inconsistent
    and see different subsets of data.

    Suppose an iterator is opened from a storage, and slightly later a
    second iterator is also opened. The two iterators may produce very
    different data, and attempting to compare them or treat the later
    one as a strict superset of the former will break in a
    history-free storage.

    For both history-free and history-preserving databases, the second
    iterator can include transactions not in the first one.

    In a history-preserving database, absent a pack, that's fine: the
    second iterator will be a strict superset of the first one.

    In a history-free database, however, when an object is changed, it is
    effectively *deleted* from its old transaction; its new state is
    only visible as a record in its new transaction.

    That's also fine: the second iterator will include that new state
    (eventually).

    Where it's not fine is when the `stop` parameter is used. **HF
    RelStorage is only fully consistent as-of the most recent
    committed transaction visible to any given connection.**

    Suppose the first iterator is opened and returns transactions 1,
    2, 3, 4, 5, where 5 is now the highest visible transaction at the
    time it is opened. In transaction 5, the state of object A
    references object C.

    Open a second iterator, which lets the view of the database move
    forward. This time, there's a new transaction, 6: the state of A
    has changed, but still contains a reference to C. If you pass
    `stop=5`, though, you won't see *any* state for A because it's
    only found in transaction 6. If A was the only reference to object
    C, object C now appears to be garbage.

    This comes up in ``zc.zodbdgc``, which does exactly that sequence by
    default, or if you pass ``--days <non-zero>``. If you pass ``--days
    0``, it only uses one iterator, but it still passes an arbitrary
    ``stop`` parameter, making that unsafe as well (it tries to use the
    current time as the value to stop at, in order to view all
    committed transactions consistently as-of this date, but that's
    insufficient and there could be committed transactions more recent
    than that; consider clock differences among machines).

    This causes data loss.

    To avoid this, pass in the storage's load cursor consistently. Only if
    the storage has been `sync()` or committed a transaction would the
    two differ.
    """
    __slots__ = ()


class RelStorageTransactionRecord(TransactionRecord):

    def __init__(self, trans_iter, tid_int, user, desc, ext, packed):
        self._trans_iter = trans_iter
        self._tid_int = tid_int
        tid = int64_to_8bytes(tid_int)
        status = 'p' if packed else ' '
        user = user or b''
        description = desc or b''
        if ext:
            extension = loads(ext)
        else:
            extension = {}

        TransactionRecord.__init__(self, tid, status, user, description, extension)

    def __iter__(self):
        return TransactionRecordIterator(self)

    def __repr__(self):
        return '<%s at %x tid=%d status=%r user=%r description=%r>' % (
            self.__class__.__name__,
            id(self),
            self._tid_int,
            self.status,
            self.user,
            self.description
        )

class TransactionRecordIterator(object):
    """Iterate over the objects in a transaction."""

    __slots__ = (
        'tid',
        '_records',
        '_index',
    )

    def __init__(self, record):
        # type: (RelStorageTransactionRecord) -> None

        cursor = record._trans_iter._cursor
        adapter = record._trans_iter._adapter
        tid_int = record._tid_int
        self.tid = record.tid
        self._records = list(adapter.dbiter.iter_objects(cursor, tid_int))
        self._index = 0

    def __iter__(self):
        return self

    def __len__(self):
        return len(self._records)

    def __getitem__(self, n):
        self._index = n
        return next(self)

    def next(self):
        if self._index >= len(self._records):
            raise StorageStopIteration()
        params = self._records[self._index]
        res = Record(self.tid, *params)
        self._index += 1
        return res

    __next__ = next


class Record(DataRecord):
    """An object state in a transaction"""

    def __init__(self, tid, oid_int, data):
        DataRecord.__init__(self, int64_to_8bytes(oid_int), tid, data, None)



##############################################################################
#
# Copyright (c) 2008 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
import sys
import os
from setuptools import setup
from setuptools import find_packages
from setuptools import Extension

# Based on code from
# http://cython.readthedocs.io/en/latest/src/reference/compilation.html#distributing-cython-modules
def _dummy_cythonize(extensions, **_kwargs):
    for extension in extensions:
        sources = []
        for sfile in extension.sources:
            path, ext = os.path.splitext(sfile)
            if ext in ('.pyx', '.py'):
                ext = '.cpp' if extension.language == 'c++' else '.c'
                sfile = path + ext
            sources.append(sfile)
        extension.sources[:] = sources
    return extensions

try:
    from Cython.Build import cythonize
except ImportError:
    # The generated source files had better already exist.
    cythonize = _dummy_cythonize


def read_file(*path):
    base_dir = os.path.dirname(__file__)
    file_path = (base_dir, ) + tuple(path)
    with open(os.path.join(*file_path)) as f:
        result = f.read()
    return result

VERSION = read_file('version.txt').strip()
PYPY = hasattr(sys, 'pypy_version_info')
WINDOWS = sys.platform.startswith("win")
PY3 = sys.version_info[0] == 3
PY2 = not PY3

memcache_require = [
    'pylibmc; platform_python_implementation=="CPython" and sys_platform != "win32"',
    'python-memcached; platform_python_implementation=="PyPy" or sys_platform == "win32"',
]

tests_require = [
    'mock ; python_version == "2.7"',
    # random2 is a forward port of python 2's random to
    # python 3. Our test_cache_stats (inherited from ZEO)
    # needs that. Without it, the tests can work on Py2 but
    # not Py3.
    'random2',
    'zope.testing',
    'ZODB [test]',
    # We have the aforementioned test_cache_stats from ZEO,
    # which obviously needs a ZEO dependency. But on Python 2.7,
    # ZEO depends on trollius, which is deprecated and can't be installed on
    # PyPy 2.7 on Windows, so don't install it.
    'ZEO >= 5.2; python_version > "2.7"',
    'zc.zlibstorage',
    'zope.testrunner',
    'nti.testing',
    'gevent >= 1.5a1; sys_platform != "win32"',
    'pyperf',
    # Versions of PyPy2 prior to 7.4 (maybe?) are incompatible with
    # psutil >= 5.6.4.
    # https://github.com/giampaolo/psutil/issues/1659
    'psutil; platform_python_implementation=="CPython" or python_version!="2.7"',
    'psutil == 5.6.3; platform_python_implementation=="PyPy" and python_version=="2.7"',
] + memcache_require


setup(
    name="RelStorage",
    version=VERSION,
    author="Shane Hathaway with Zope Foundation and Contributors",
    author_email="shane@willowrise.com",
    maintainer="Jason Madden",
    maintainer_email="jason@nextthought.com",
    url="https://relstorage.readthedocs.io/",
    project_urls={
        'Bug Tracker': 'https://github.com/zodb/relstorage/issues',
        'Source Code': 'https://github.com/zodb/relstorage/',
        'Documentation': 'https://relstorage.readthedocs.io',
    },
    keywords="ZODB SQL RDBMS MySQL PostgreSQL Oracle",
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    license="ZPL 2.1",
    platforms=["any"],
    description="A backend for ZODB that stores pickles in a relational database.",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Zope Public License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Database",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: Unix",
        "Development Status :: 4 - Beta",
    ],
    long_description=read_file("README.rst"),
    # We cannot be used from an archive. ZConfig can't see
    # our component.xml, and we rely on being able to use __file__ to
    # list and locate auxiliary files (e.g., schema.py finds SQL files)
    zip_safe=False,
    setup_requires=[
    ],
    install_requires=[
        'perfmetrics >= 3.0.0',
        'zope.interface',
        'zope.dottedname',
        'zc.lockfile',
        'BTrees >= 4.7.2', # unsigned btrees in 4.7+; correct errors in 4.7.2
        # These are the versions we're testing against. ZODB 5.2.2 is
        # when checkSecure() went away and IMVCCAfterCompletionStorage
        # was added, and 5.1.2 is when Connection.new_oid was added
        # (https://github.com/zopefoundation/ZODB/issues/139)
        'ZODB >= 5.5',
        # We directly use this, and its a transient dep of ZODB.
        # version 2.0 is where things became text, and 2.1 partly
        # relaxed those requirements.
        'transaction >= 2.4.0',
    ],
    ext_modules=cythonize(
        [
            Extension(
                name="relstorage.cache.cache",
                language="c++",
                sources=[
                    'src/relstorage/cache/cache.pyx',
                    'src/relstorage/cache/c_cache.cpp',
                ],
                include_dirs=['include'],
                # https://docs.microsoft.com/en-us/cpp/build/reference/eh-exception-handling-model
                # /EHsc: `s`: enable stack unwinding, catch C++ exceptions in catch(...)
                #        `c`: extern C functions never throw C++ exceptions.
                # XXX: Why this flag or Python 2/Windows? /EHa (catch SEH and standard) seems better.
                extra_compile_args=["/EHsc"] if WINDOWS and PY2 else [],
            ),

        ],
        annotate=True,
        compiler_directives={
            'language_level': '3str',
            'always_allow_keywords': False,
            'infer_types': False,
            'nonecheck': False,
        },
    ),
    tests_require=tests_require,
    extras_require={
        # We previously used MySQL-python (C impl) on CPython 2.7,
        # because it had the longest history. However, it is no longer
        # maintained, so we switch to its maintained fork mysqlclient;
        # this is consistent with Python 3. The best option for PyPy
        # is PyMySQL because MySQL-python doesn't support it (and
        # binary drivers like that tend to be slow). Although both
        # PyMySQL and mysqlclient support Python 3, use mysqlclient on
        # Python 3 because it's a binary driver and *probably* faster
        # for CPython; it requires some minor code changes to support,
        # so be sure to test this configuration. mysqlclient doesn't
        # compile on windows for Python 2.7 (it doesn't support the
        # old Visual C) and is hard to compile for anything newer; it sometimes
        # has wheels but not always, so don't suggest it.
        #
        # Note that mysqlclient 2.0.0 has a crashing bug:
        # https://github.com/PyMySQL/mysqlclient-python/issues/435


        # pylint:disable=line-too-long
        'mysql:platform_python_implementation=="CPython" and (sys_platform != "win32")': [
            'mysqlclient >= 1.4, != 2.0.0',
        ],
        'mysql:platform_python_implementation=="PyPy" or (sys_platform == "win32")': [
            'PyMySQL>=0.6.6',
        ],
        # Notes on psycopg2: In 2.8, they stopped distributing full
        # binary wheels under that name. For that, you have to use
        # psycopg2-binary (which in 2.8.3 appears to be compiled with
        # the libpq from postgres 11). But that's not recommended for
        # production usage because it can have conflicts with other libraries,
        # and the authors specifically request that other modules not depend on
        # psycopg2-binary.
        # See http://initd.org/psycopg/docs/install.html#binary-packages
        'postgresql: platform_python_implementation == "CPython"': [
            # 2.4.1+ is required for proper bytea handling;
            # 2.6+ is needed for 64-bit lobject support;
            # 2.7+ is needed for Python 3.7 support and PostgreSQL 10+;
            # 2.7.6+ is needed for PostgreSQL 11;
            # 2.8 is needed for conn.info
            'psycopg2 >= 2.8.3',
        ],
        'postgresql: platform_python_implementation == "PyPy"': [
            # 2.8.0+ is needed for Python 3.7
            'psycopg2cffi >= 2.8.1',
        ],
        'oracle': [
            # 5.2 added support for LOBs larger than 4GB.
            # 6.0 lets lobs be used across DB fetches, uses
            # temp lob caching.
            # 6.2 lets lobs be bound directly to a cursor, and
            # 7.0 lets lobs be set directly from bytes, without an intermediate
            # temporary lob.
            'cx_Oracle>=6.0'
        ],
        'sqlite': [],
        'sqlite3': [],
        'memcache': memcache_require,
        'test': tests_require,
        'docs': [
            'sphinx',
            'sphinxcontrib-programoutput',
            'repoze.sphinx.autointerface',
            'sphinx_rtd_theme',
        ],
        'all_tested_drivers': [
            # Install all the supported drivers for the platform.
            # Spread them out across the versions to not load any one
            # up too heavy for better parallelism.

            # First, mysql
            # pymysql on 3.6 on all platforms. We get coverage data from Travis,
            # and we get 2.7 from PyPy and Windows.
            'PyMySQL >= 0.6.6; python_version == "3.6" or platform_python_implementation == "PyPy" or sys_platform == "win32"',
            # mysqlclient (binary) on all CPythons. It's the default,
            # except on old Windows. We get coverage from Travis.
            'mysqlclient >= 1.4,!=2.0.0;platform_python_implementation=="CPython" and (sys_platform != "win32")',
            # mysql-connector-python on Python 3.7 for coverage on Travis and ensuring it works
            # on Windows, and PyPy for testing there, since it's one of two pure-python versions.
            'mysql-connector-python >= 8.0.16; python_version == "3.7" or platform_python_implementation == "PyPy"',

            # postgresql
            # pure-python
            # pg8000 on Python 2.7 or PyPy. We get coverage from Travis, and we also
            # get Windows.
            'pg8000 >= 1.11.0; python_version == "2.7" or platform_python_implementation == "PyPy"',
            # CFFI, runs on all implementations.
            # We get coverage from 3.5 on Travis and verification it works on Windows from Travis.
            # We get 2.7 testing from PyPy on Travis.
            'psycopg2cffi >= 2.7.4; python_version == "3.5" or platform_python_implementation == "PyPy"',
            # Psycopg2 on all CPython, it's the default
            'psycopg2 >= 2.8.3; platform_python_implementation == "CPython"',
        ],
    },
    entry_points={
        'console_scripts': [
            'zodbconvert = relstorage.zodbconvert:main',
            'zodbpack = relstorage.zodbpack:main',
        ],
        'zodburi.resolvers': [
            'postgres = relstorage.zodburi_resolver:postgresql_resolver',
            'mysql = relstorage.zodburi_resolver:mysql_resolver',
            'oracle = relstorage.zodburi_resolver:oracle_resolver',
            'sqlite = relstorage.zodburi_resolver:sqlite_resolver',
        ],
        'gevent.plugins.monkey.did_patch_builtins': [
            'psycopg2 = relstorage.adapters.postgresql.drivers.psycopg2:_gevent_did_patch',
        ],
    },
)



##############################################################################
#
# Copyright (c) 2009 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""Pack/Undo implementations.
"""

from __future__ import absolute_import

from ..packundo import HistoryFreePackUndo
from ..packundo import HistoryPreservingPackUndo
from ..schema import Schema

class _LockStmt(object):
    # 8.0 supports 'FOR SHARE' but before that we have
    # this.
    _lock_for_share = 'LOCK IN SHARE MODE'

class MySQLHistoryPreservingPackUndo(_LockStmt, HistoryPreservingPackUndo):

    # Previously we needed to work around a MySQL performance bug by
    # avoiding an expensive subquery.
    #
    # See:
    #      http://mail.zope.org/pipermail/zodb-dev/2008-May/011880.html
    #      http://bugs.mysql.com/bug.php?id=28257
    #
    # However, this was fixed in 5.6.
    _script_create_temp_pack_visit = """
        CREATE TEMPORARY TABLE temp_pack_visit (
            zoid BIGINT UNSIGNED NOT NULL PRIMARY KEY,
            keep_tid BIGINT UNSIGNED NOT NULL
        );
        CREATE INDEX temp_pack_keep_tid ON temp_pack_visit (keep_tid);
        """

    # It was once purported here that "MySQL optimizes deletion far
    # better when using a (USING) join syntax", a so-called
    # "multi-table" delete; indeed, as of 5.7 and 8.0, the documentation still maintains
    # that certain optimizations aren't used in DELETE with subqueries and councils
    # USING instead. However: "You cannot use ORDER BY or LIMIT
    # in a multiple-table DELETE" and "If you use a multiple-table
    # DELETE statement involving InnoDB tables for which there are
    # foreign key constraints, the MySQL optimizer might process
    # tables in an order that differs from that of their parent/child
    # relationship. In this case, the statement fails and rolls back.
    # Instead, you should delete from a single table and rely on the
    # ON DELETE capabilities that InnoDB provides to cause the other
    # tables to be modified accordingly."
    # (https://dev.mysql.com/doc/refman/5.7/en/delete.html; The same
    # goes for 8.0). In 8.0 we could potentially use a WITH clause.
    #
    # ORDER BY is important to lock rows in the same order as
    # transaction commits do, so that we don't deadlock with
    # commits, which also lock rows.
    _script_pack_current_object = """
    DELETE FROM current_object
    WHERE zoid IN (
        SELECT zoid
        FROM pack_state
        WHERE pack_state.tid = %(tid)s
    )
    AND tid = %(tid)s
    ORDER BY zoid
    """

    _script_pack_object_state = """
    DELETE FROM object_state
    WHERE zoid IN (
        SELECT zoid
        FROM pack_state
        WHERE pack_state.tid = %(tid)s
    )
    AND tid = %(tid)s
    ORDER BY zoid
    """

    _script_pack_object_ref = """
        DELETE FROM object_refs_added
        USING object_refs_added
            JOIN transaction USING (tid)
        WHERE transaction.is_empty = true;

        DELETE FROM object_ref
        USING object_ref
            JOIN transaction USING (tid)
        WHERE transaction.is_empty = true
        """

    _script_create_temp_undo = """
        CREATE TEMPORARY TABLE temp_undo (
            zoid BIGINT UNSIGNED NOT NULL PRIMARY KEY,
            prev_tid BIGINT UNSIGNED NOT NULL
        );
    """

    # pylint:disable=singleton-comparison
    _delete_empty_transactions_batch_query = Schema.transaction.delete(
    ).where(
        Schema.transaction.c.packed == True
    ).and_(
        Schema.transaction.c.is_empty == True
    ).limit(1000)


class MySQLHistoryFreePackUndo(_LockStmt, HistoryFreePackUndo):
    pass



# -*- coding: utf-8 -*-

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from relstorage.tests import TestCase
from relstorage.adapters.tests.test_drivers import IDBDriverSupportsCriticalTestMixin
from ..import psycopg2



class GeventPsycopg2Driver(psycopg2.GeventPsycopg2Driver):
    GeventConnection = None


    def _create_connection(self, mod, *extra_slots):
        cls = type(self)
        if cls.GeventConnection is None:
            base = super(GeventPsycopg2Driver, self)._create_connection(mod, *extra_slots)

            class GeventConnection(base):
                def __init__(self, *args, **kwargs): # pylint:disable=super-init-not-called
                    self.connect_args = args
                    self.conect_kwargs = kwargs
            cls.GeventConnection = GeventConnection
        return cls.GeventConnection

    def _check_wait_callback(self):
        return True

class TestGeventPsycopg2(IDBDriverSupportsCriticalTestMixin,
                         TestCase):

    def _makeOne(self):
        try:
            return GeventPsycopg2Driver()
        except ImportError as e:
            self.skipTest(e)

    def test_sync_ends_critical_phase(self):
        driver = self._makeOne()
        conn = driver.connect()
        driver.enter_critical_phase_until_transaction_end(conn, None)
        self.assertTrue(conn.is_in_critical_phase())
        driver.sync_status_after_hidden_commit(conn)
        self.assertFalse(driver.is_in_critical_phase(conn, None))

    def test_execute_multiple_statement_with_hidden_commit(self):
        driver = self._makeOne()
        conn = driver.connect()
        driver.enter_critical_phase_until_transaction_end(conn, None)
        class MockCursor(object):
            def execute(self, *_args):
                pass

        self.assertTrue(conn.is_in_critical_phase())
        driver.execute_multiple_statement_with_hidden_commit(
            conn, MockCursor(), 'SELECT; COMMIT;', ()
        )
        self.assertFalse(driver.is_in_critical_phase(conn, None))



# -*- coding: utf-8 -*-
# distutils: language = c++
# cython: auto_pickle=False,embedsignature=True,always_allow_keywords=False,infer_types=True

from libcpp.vector cimport vector
from libcpp cimport bool


cdef extern from *:
    ctypedef signed long int64_t
    ctypedef signed long uint64_t

cdef extern from * namespace "boost":
    T& move[T](T&)


cdef extern from "c_cache.h" namespace "relstorage::cache":
    ctypedef int64_t TID_t
    ctypedef int64_t OID_t

    ctypedef enum generation_num:
        GEN_UNKNOWN,
        GEN_EDEN,
        GEN_PROTECTED,
        GEN_PROBATION


    cdef cppclass ICacheEntry:
        OID_t key
        size_t weight() except +
        size_t frequency
        bint in_cache() except +
        vector[TID_t] all_tids() except +
        # Memory management
        bool can_delete() except +
        const T* Py_use[T]()
        bool Py_release() except + # Throwing during dev because of assertions


    cdef cppclass SVCacheEntry(ICacheEntry):
        TID_t tid() const
        bint frozen() const
        # Using -1 for None
        object as_object()
        size_t size()
        bint operator==(SVCacheEntry&)


    cdef cppclass MVCacheEntry(ICacheEntry):
        SVCacheEntry* copy_newest_entry() except +
        TID_t newest_tid()


    cdef cppclass TempCacheFiller:
        TempCacheFiller()
        TempCacheFiller(size_t number_nodes) except +
        void set_once_at(size_t, OID_t, TID_t, object, bool frozen, int frequency)


    cdef cppclass Generation:
        cppclass iterator:
            bool operator==(iterator)
            bool operator!=(iterator)
            const ICacheEntry& operator*()
            iterator& operator++()
        size_t size() const
        size_t sum_weights() const
        size_t max_weight()
        const generation_num generation
        bool empty()
        iterator begin()
        iterator end()

    cdef cppclass ProposedCacheEntry:
        ProposedCacheEntry()
        ProposedCacheEntry(OID_t, TID_t, object)


    cdef cppclass Cache:
        cppclass iterator:
            bool operator==(iterator)
            bool operator!=(iterator)
            ICacheEntry& operator*()
            iterator& operator++()
            # Sadly this is not supported:
            # ICacheEntry& operator->()

        Generation ring_eden
        Generation ring_protected
        Generation ring_probation
        Cache()
        Cache(size_t eden, size_t protected, size_t probation) except +
        void __del__() except +
        void resize(size_t, size_t, size_t)
        size_t max_weight()
        void add_to_eden(ProposedCacheEntry) except +
        void store_and_make_MRU(ProposedCacheEntry) except +
        void delitem(OID_t key) except +
        void delitem(OID_t key, TID_t tid) except +
        void freeze(OID_t key, TID_t tid) except +
        bool contains(OID_t key)
        void age_frequencies()
        ICacheEntry* get(OID_t key)
        SVCacheEntry* get(OID_t, TID_t)
        SVCacheEntry* peek(OID_t, TID_t)
        size_t size()
        size_t weight()
        vector[OID_t] add_many(TempCacheFiller&) except +

        iterator begin()
        iterator end()


# Local Variables:
# flycheck-cython-cplus: t
# End:



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
psycpg2-like lobject support using the SQL interface,
for drivers and modes that don't support it natively.

For example, pg8000 provides no LOB support, and psycopg2 when in
asynchronous mode has no LOB support. This package provides just enough
functionality for our purposes. Activate it by deriving a class from
the driver's Connection class and this module's :class:`LobConnectionMixin`.
"""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division

__all__ = [
    'LobConnectionMixin',
]


class _WriteBlob(object):
    closed = False

    def __init__(self, conn, binary):
        self._binary = binary
        self._cursor = conn.cursor()
        self._offset = 0
        try:
            self._cursor.execute("SELECT lo_creat(-1)")
            row = self._cursor.fetchone()
            self.oid = row[0]
        except:
            self._cursor.close()
            raise

    def close(self):
        self._cursor.close()
        self.closed = True

    def write(self, data):
        self._cursor.execute("SELECT lo_put(%(oid)s, %(off)s, %(data)s)",
                             {'oid': self.oid, 'off': self._offset, 'data': self._binary(data)})
        self._offset += len(data)
        return len(data)

class _UploadBlob(object):
    closed = False
    fetch_size = 1024 * 1024 * 9

    def __init__(self, conn, new_file, binary):
        blob = _WriteBlob(conn, binary)
        self.oid = blob.oid
        try:
            with open(new_file, 'rb') as f:
                while 1:
                    data = f.read(self.fetch_size)
                    if not data:
                        break
                    blob.write(data)
        finally:
            blob.close()

    def close(self):
        self.closed = True

class _ReadBlob(object):
    closed = False
    fetch_size = 1024 * 1024 * 9

    def __init__(self, conn, oid):
        self._cursor = conn.cursor()
        self.oid = oid
        self.offset = 0

    def export(self, filename):
        with open(filename, 'wb') as f:
            while 1:
                data = self.read(self.fetch_size)
                if not data:
                    break
                f.write(data)
        self.close()

    def read(self, size):
        self._cursor.execute("SELECT lo_get(%(oid)s, %(off)s, %(cnt)s)",
                             {'oid': self.oid, 'off': self.offset, 'cnt': size})
        row = self._cursor.fetchone()
        data = row[0]
        self.offset += len(data)
        return data

    def close(self):
        self._cursor.close()
        self.closed = True



class LobConnectionMixin(object):

    #: The driver module's ``Binary`` converter object.
    RSDriverBinary = None

    def lobject(self, oid=0, mode='', new_oid=0, new_file=None):
        if oid == 0 and new_oid == 0 and mode == 'wb':
            if new_file:
                # Upload the whole file right now.
                return _UploadBlob(self, new_file, self.RSDriverBinary)
            return _WriteBlob(self, self.RSDriverBinary)
        if oid != 0 and mode == 'rb':
            return _ReadBlob(self, oid)
        raise AssertionError("Unsupported params", dict(locals()))



##############################################################################
#
# Copyright (c) 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################
"""
Batch table row insert/delete support.
"""

from __future__ import absolute_import


from ..batch import RowBatcher

class PostgreSQLRowBatcher(RowBatcher):
    """
    Applies array operations to DELETE and SELECT
    for single column filters.
    """

    def _make_single_column_query(self, command, table,
                                  filter_column, filter_value,
                                  rows_need_flattened):
        stmt = "%s FROM %s WHERE %s = ANY (%s)" % (
            command, table, filter_column, self.delete_placeholder
        )
        # We only pass a single parameter, and it doesn't need further
        # flattening.
        # It does have to be an actual list, though
        params = filter_value
        if rows_need_flattened:
            # This is guaranteed to return a list
            params = self._flatten_params(params)
        elif not isinstance(params, list):
            params = list(params)
        return stmt, (params,), False



# -*- coding: utf-8 -*-
##############################################################################
#
# Copyright (c) 2008, 2019 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#
##############################################################################

"""
Implementation of the oid allocation algorithm.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from ZODB.POSException import ReadOnlyError
from ZODB.utils import p64 as int64_to_8bytes

from zope.interface import implementer

from .interfaces import IStaleAware

logger = __import__('logging').getLogger(__name__)

class AbstractOIDs(object):

    __slots__ = (
    )

    def stale(self, ex):
        return StaleOIDs(ex, self)

    def no_longer_stale(self):
        return self

    def new_oid(self, store_connection_pool, commit_in_progress):
        raise NotImplementedError

    def set_min_oid(self, store_connection, max_observed_oid):
        raise NotImplementedError

class OIDs(AbstractOIDs):

    def __init__(self, oidallocator):
        # From largest to smallest: [16, 15, 14, ..., 1]
        self.preallocated_oids = [] # type: list
        # The maximum OID we've handed out (or that has been observed)
        # A value of 0 is not legal for the oidallocator to produce.
        self.max_allocated_oid = 0 # type: int
        self.oidallocator = oidallocator
        if hasattr(oidallocator, 'new_oids_no_cursor'):
            self.__preallocate_oids = self.__preallocate_oids_no_cursor

    def set_min_oid(self, store_connection, max_observed_oid):
        """
        Ensure that the next oid we produce is greater than *max_observed_oid*.

        Must be done in a transaction while the store connection is usable.
        """
        if max_observed_oid > self.max_allocated_oid:
            # They saw one from outside of us that's greater than what
            # we've allocated. We could be a brand new object that's
            # never allocated an OID before (i.e., we're unused, or
            # we've only done loads); or, we could be copying
            # transactions from an external storage.

            # Set it in the database for everyone.
            self.oidallocator.set_min_oid(store_connection.cursor,
                                          max_observed_oid)
            # Then, set it in the storage for this thread
            # so we don't have to keep doing this if it only ever
            # updates existing objects.
            # NOTE: This is a non-transactional change to the our state.
            # That's OK, though, as the underlying sequence for OIDs we allocate
            # is also non-transactional.
            self.max_allocated_oid = max_observed_oid
            # Discard any preallocated oids that are less than this; they're not
            # safe to use in the most general case. (In the typical case they probably
            # are.)
            preallocated_oids = self.preallocated_oids
            while preallocated_oids and preallocated_oids[-1] < max_observed_oid:
                preallocated_oids.pop()

    def new_oid(self, store_connection_pool, commit_in_progress):
        # Prior to ZODB 5.1.2, this method was actually called on the
        # storage object of the DB, not the instance storage object of
        # a Connection. This meant that this method (and the oid
        # cache) was shared among all connections using a database and
        # was called outside of a transaction (starting its own
        # long-running transaction).

        # The DB.new_oid() method still exists, but shouldn't be used;
        # In the past, because that storage was never in a transaction, that meant that
        # we could open a database transaction and never close it. RelStorage 3.3 and
        # store connection pooling fixes this.

        # Connection.new_oid() can be called at just about any time
        # thanks to the Connection.add() API, which clients can use
        # at any time (typically before commit begins, but it's possible to
        # add() objects from a ``__getstate__`` method).
        #
        # Thus we may or may not have a store connection already open;
        # if we do, we can't restart it or drop it.
        if not self.preallocated_oids:
            self.__preallocate_oids(store_connection_pool, commit_in_progress)
            # OIDs are monotonic, always increasing. It should never
            # go down or return equal to what we've already seen.
            self.max_allocated_oid = max(self.preallocated_oids[0], self.max_allocated_oid)

        oid_int = self.preallocated_oids.pop()
        return int64_to_8bytes(oid_int)

    def __preallocate_oids(self, store_connection_pool, commit_in_progress): # pylint:disable=method-hidden
        with store_connection_pool.borrowing(commit=True) as store_connection:
            self.preallocated_oids = store_connection.call(
                self.__new_oid_callback,
                can_reconnect=not commit_in_progress
            )

    def __preallocate_oids_no_cursor(self, _store_connection, commit_in_progress):# pylint:disable=unused-argument
        self.preallocated_oids = self.oidallocator.new_oids_no_cursor()

    def __new_oid_callback(self, _store_conn, store_cursor, _fresh_connection):
        return self.oidallocator.new_oids(store_cursor)


@implementer(IStaleAware)
class ReadOnlyOIDs(AbstractOIDs):

    __slots__ = (
    )

    def new_oid(self, store_connection_pool, commit_in_progress):
        raise ReadOnlyError

    def set_min_oid(self, store_connection, max_observed_oid):
        raise ReadOnlyError

@implementer(IStaleAware)
class StaleOIDs(AbstractOIDs):

    __slots__ = (
        'stale_error',
        'previous',
    )

    def __init__(self, stale_error, previous):
        self.stale_error = stale_error
        self.previous = previous

    def no_longer_stale(self):
        return self.previous

    def stale(self, ex):
        return self

    def new_oid(self, store_connection_pool, commit_in_progress):
        raise self.stale_error

    def set_min_oid(self, store_connection, max_observed_oid):
        raise self.stale_error
